"""
CPython 3.2 bytecode opcodes to be used as a base for other opcodes including 3.2.

This is used in bytecode disassembly among other things. This is
similar to the opcodes in Python's opcode.py library.

If this file changes the other opcode files may have to a adjusted accordingly.
"""

from xdis.opcodes.base import (
    compare_op, const_op,
    def_op, format_extended_arg,
    free_op, jabs_op, jrel_op,
    local_op, name_op, nargs_op,
    varargs_op
    )

l = locals()

# FIXME: DRY with opcode2x.py

hascompare   = []
hascondition = [] # conditional operator; has jump offset
hasconst     = []
hasfree      = []
hasjabs      = []
hasjrel      = []
haslocal     = []
hasname      = []
hasnargs     = []  # For function-like calls
hasvargs     = []  # Similar but for operators BUILD_xxx
nofollow     = []  # Instruction doesn't fall to the next opcode

# opmap[opcode_name] => opcode_number
opmap = {}

# opcode[i] => opcode name
opname = [''] * 256

# oppush[op] => number of stack entries pushed
# -9 means handle special. Note his forces oppush[i] - oppop[i] negative
oppush = [0] * 256

# -1 means handle special.
# oppop[op] => number of stack entries popped
oppop  = [0] * 256

for op in range(256): opname[op] = '<%r>' % (op,)
del op

# Instruction opcodes for compiled code
# Blank lines correspond to available opcodes

#          OP NAME            OPCODE POP PUSH
#--------------------------------------------
def_op(l, 'STOP_CODE',             0,  0,  0, fallthrough=False)
def_op(l, 'POP_TOP',               1,  1,  0)
def_op(l, 'ROT_TWO',               2,  2,  2)
def_op(l, 'ROT_THREE',             3,  3,  3)
def_op(l, 'DUP_TOP',               4,  0,  1)

# Python 3.2+
def_op(l, 'DUP_TOP_TWO',           5,  0,  2)

def_op(l, 'NOP', 9)
def_op(l, 'UNARY_POSITIVE',       10,  1,  1)
def_op(l, 'UNARY_NEGATIVE',       11,  1,  1)
def_op(l, 'UNARY_NOT',            12,  1,  1)

def_op(l, 'UNARY_INVERT',         15,  1,  1)

def_op(l, 'BINARY_POWER',         19,  2,  1)
def_op(l, 'BINARY_MULTIPLY',      20,  2,  1)

def_op(l, 'BINARY_MODULO',        22,  2,  1)
def_op(l, 'BINARY_ADD',           23,  2,  1)
def_op(l, 'BINARY_SUBTRACT',      24,  2,  1)
def_op(l, 'BINARY_SUBSCR',        25,  2,  1)
def_op(l, 'BINARY_FLOOR_DIVIDE',  26,  2,  1)
def_op(l, 'BINARY_TRUE_DIVIDE',   27,  2,  1)
def_op(l, 'INPLACE_FLOOR_DIVIDE', 28,  2,  1)
def_op(l, 'INPLACE_TRUE_DIVIDE',  29,  2,  1)

# Gone from Python 3 are Python2's
# SLICE+0 .. SLICE+3
# STORE_SLICE+0 .. STORE_SLICE+3
# DELETE_SLICE+0 .. DELETE_SLICE+3

def_op(l, 'STORE_MAP',            54,  3,  1)
def_op(l, 'INPLACE_ADD',          55,  2,  1)
def_op(l, 'INPLACE_SUBTRACT',     56,  2,  1)
def_op(l, 'INPLACE_MULTIPLY',     57,  2,  1)

def_op(l, 'INPLACE_MODULO',       59,  2,  1)
def_op(l, 'STORE_SUBSCR',         60,  3,  0) # Implements TOS1[TOS] = TOS2.
def_op(l, 'DELETE_SUBSCR',        61,  2,  0) # Implements del TOS1[TOS].
def_op(l, 'BINARY_LSHIFT',        62,  2,  1)
def_op(l, 'BINARY_RSHIFT',        63,  2,  1)
def_op(l, 'BINARY_AND',           64,  2,  1)
def_op(l, 'BINARY_XOR',           65,  2,  1)
def_op(l, 'BINARY_OR',            66,  2,  1)
def_op(l, 'INPLACE_POWER',        67,  2,  1)
def_op(l, 'GET_ITER',             68,  1,  1)
def_op(l, 'STORE_LOCALS',         69,  1,  0)

def_op(l, 'PRINT_EXPR',           70,  1,  0)
def_op(l, 'LOAD_BUILD_CLASS',     71,  0,  1)

# Python3 drops/changes:
#  def_op(l, 'PRINT_ITEM', 71)
#  def_op(l, 'PRINT_NEWLINE', 72)
#  def_op(l, 'PRINT_ITEM_TO', 73)
#  def_op(l, 'PRINT_NEWLINE_TO', 74)

def_op(l, 'INPLACE_LSHIFT',       75,  2,  1)
def_op(l, 'INPLACE_RSHIFT',       76,  2,  1)
def_op(l, 'INPLACE_AND',          77,  2,  1)
def_op(l, 'INPLACE_XOR',          78,  2,  1)
def_op(l, 'INPLACE_OR',           79,  2,  1)
def_op(l, 'BREAK_LOOP',           80,  0,  0)
def_op(l, 'WITH_CLEANUP',         81, -1, -9) # Cleans up the stack when a with statement
                                              # block exits.  Handle stack special

def_op(l, 'RETURN_VALUE',         83,  1,  0, fallthrough=False)
def_op(l, 'IMPORT_STAR',          84,  1,  0)

def_op(l, 'YIELD_VALUE',          86,  1,  1)
def_op(l, 'POP_BLOCK',            87,  0,  0)
def_op(l, 'END_FINALLY',          88,  1,  0)
def_op(l, 'POP_EXCEPT',           89,  1, -1)

HAVE_ARGUMENT = 90              # Opcodes from here have an argument:

name_op(l, 'STORE_NAME',            90,  1,  0)   # Operand is in name list
name_op(l, 'DELETE_NAME',           91,  0,  0)   # ""
varargs_op(l, 'UNPACK_SEQUENCE',    92, -1,  1)   # Number of tuple items
jrel_op(l,    'FOR_ITER',           93,  1, -9)

def_op(l,  'UNPACK_EX',             94, -1, -9)   # assignment with a starred target
                                                  # argument has a count
name_op(l, 'STORE_ATTR',            95,  2,  0)   # Operand is in name list
name_op(l, 'DELETE_ATTR',           96,  1,  0)   # ""
name_op(l, 'STORE_GLOBAL',          97,  1,  0)   # ""
name_op(l, 'DELETE_GLOBAL',         98,  0,  0)   # ""

# Python 2's DUP_TOPX is gone starting in Python 3.2

const_op(l,   'LOAD_CONST',        100,  0,  1)  # Operand is in const list
name_op(l,    'LOAD_NAME',         101,  0,  1)  # Operand is in name list
varargs_op(l, 'BUILD_TUPLE',       102, -1,  1)  # TOS is count of tuple items
varargs_op(l, 'BUILD_LIST',        103, -1,  1)  # TOS is count of list items
varargs_op(l, 'BUILD_SET',         104, -1,  1)  # TOS is count of set items
varargs_op(l, 'BUILD_MAP',         105, -1,  1)  # TOS is count of kwarg items
name_op(l, 'LOAD_ATTR',            106,  1,  1)  # Operand is in name list
compare_op(l, 'COMPARE_OP',        107,  2,  1)  # Comparison operator
name_op(l, 'IMPORT_NAME',          108,  1,  1)  # Operand is in name list
name_op(l, 'IMPORT_FROM',          109,  0,  1)  # Operand is in name list

jrel_op(l, 'JUMP_FORWARD',         110,  0,  0)  # Number of bytes to skip
jabs_op(l, 'JUMP_IF_FALSE_OR_POP', 111, conditional=True)  # Target byte offset from beginning of code
jabs_op(l, 'JUMP_IF_TRUE_OR_POP',  112, conditional=True) # ""
jabs_op(l, 'JUMP_ABSOLUTE',        113,  0,  0)  # Target byte offset from beginning of code
jabs_op(l, 'POP_JUMP_IF_FALSE',    114,  1, -9, conditional=True) # ""
jabs_op(l, 'POP_JUMP_IF_TRUE',     115,  1, -9, conditional=True) # ""

name_op(l, 'LOAD_GLOBAL',          116,  0,  1)  # Operand is in name list

jabs_op(l, 'CONTINUE_LOOP',        119,  0,  0)  # Target address
jrel_op(l, 'SETUP_LOOP',           120,  0,  0, conditional=True) # Distance to target address
jrel_op(l, 'SETUP_EXCEPT',         121,  0,  0, conditional=True)  # ""
jrel_op(l, 'SETUP_FINALLY',        122,  0,  0, conditional=True)  # ""

local_op(l, 'LOAD_FAST',           124,  0,  1)  # Local variable number
local_op(l, 'STORE_FAST',          125,  1,  0)  # Local variable number
local_op(l, 'DELETE_FAST',         126,  0,  0)  # Local variable number

def_op(l, 'RAISE_VARARGS',        130, -1,  0, fallthrough=False)
                                                 # Number of raise arguments (1, 2, or 3)
nargs_op(l, 'CALL_FUNCTION',       131, -1,  1)  # #args + (#kwargs << 8)

def_op(l, 'MAKE_FUNCTION',         132, -1,  1)  # Number of args if < 3.6
varargs_op(l, 'BUILD_SLICE',       133, -1,  1)  # Number of items

def_op(l, 'MAKE_CLOSURE',          134, -1,  1)
free_op(l, 'LOAD_CLOSURE',         135,  0,  1)
free_op(l, 'LOAD_DEREF',           136,  0,  1)
free_op(l, 'STORE_DEREF',          137,  1,  0)
free_op(l, 'DELETE_DEREF',         138,  0,  0)

nargs_op(l, 'CALL_FUNCTION_VAR',   140, -1,  1)  # #args + (#kwargs << 8)
nargs_op(l, 'CALL_FUNCTION_KW',    141, -1,  1)  # #args + (#kwargs << 8)
nargs_op(l, 'CALL_FUNCTION_VAR_KW',142, -1, 1)   # #args + (#kwargs << 8)

jrel_op(l, 'SETUP_WITH',           143,  0,  7)

def_op(l, 'LIST_APPEND',           145,  2,  1)  # Calls list.append(TOS[-i], TOS).
                                                 # Used to implement list comprehensions.
def_op(l, 'SET_ADD',               146,  1,  0)  # Calls set.add(TOS1[-i], TOS).
                                                 # Used to implement set comprehensions.
def_op(l, 'MAP_ADD',               147,  2,  1)  # Calls dict.setitem(TOS1[-i], TOS, TOS1)
                                                 # Used to implement dict comprehensions.

def_op(l, 'EXTENDED_ARG', 144)
EXTENDED_ARG = 144

def format_MAKE_FUNCTION_arg(argc):
    pos_args = argc & 0xFF
    name_default = (argc >> 8) & 0xFF
    annotate_args = (argc >> 16) & 0x7FFF
    return ("%d positional, %d name and default, %d annotations" %
            (pos_args, name_default, annotate_args))

opcode_arg_fmt = {
    'EXTENDED_ARG': format_extended_arg
}



#!/usr/bin/env python
# Copyright Hagen Fritsch, 2012, License: GPL-2.0
# Adaption and generalization for xdis use by Rocky Bernstein

# Dropbox Python bytecode decryption tool for Dropbox versions of 1.1x
# (and possibly earlier) which uses Python bytecode 2.5.

from __future__ import print_function

import types
import struct

from xdis import PYTHON3
import xdis.marsh as xmarshal
from xdis.code import Code2Compat

def rng(a, b):
    b = ((b << 13) ^ b) & 0xffffffff
    c = (b ^ (b >> 17))
    c = (c ^ (c << 5))
    return (a * 69069 + c + 0x6611CB3B) & 0xffffffff

# this is replaced by mersenne in newer versions
def get_keys(a, b):
    ka = rng(a,b)
    kb = rng(ka, a)
    kc = rng(kb, ka)
    kd = rng(kc, kb)
    ke = rng(kd, kc)
    return (kb,kc,kd,ke)

def MX(z, y, sum, key, p, e):
    return (((z >> 5 ^ y << 2) + (y >> 3 ^ z << 4)) ^ ((sum ^ y) + (key[(p & 3)^e] ^ z)))

def tea_decipher(v, key):
    """
    Tiny Decryption Algorithm decription (TEA)
    See https://en.wikipedia.org/wiki/Tiny_Encryption_Algorithm
    """
    DELTA = 0x9e3779b9
    n = len(v)
    rounds = 6 + 52//n
    sum = (rounds*DELTA)
    y = v[0]
    while sum != 0:
        e = (sum >> 2) & 3
        for p in range(n-1, -1, -1):
            z = v[(n + p - 1) % n]
            v[p] = (v[p] - MX(z, y, sum, key, p, e)) & 0xffffffff
            y = v[p]
        sum -= DELTA
    return v

def load_code(self):
    """
    Returns a Python code object like xdis.unmarshal.load_code(),
    but in we decrypt the data in self.bufstr.

    That is:
      * calculate the TEA key,
      * decrypt self.bufstr
      * create and return a Python code-object
    """
    a = self.load_int()
    b = self.load_int()
    key = get_keys(a, b)
    padsize = (b + 15) & ~0xf
    intsize = padsize/4
    data = self.bufstr[self.bufpos:self.bufpos+padsize]
    # print("%d: %d (%d=%d)" % (self.bufpos, b, padsize, len(data)))
    data = list(struct.unpack('<%dL' % intsize, data))
    tea_decipher(data, key)
    self.bufpos += padsize
    obj = xmarshal._FastUnmarshaller(struct.pack('<%dL' % intsize, *data))
    code = obj.load_code()
    co_code = patch(code.co_code)
    if PYTHON3:
        return Code2Compat(code.co_argcount, code.co_nlocals, code.co_stacksize,
                           code.co_flags,
                           co_code, code.co_consts, code.co_names, code.co_varnames,
                           code.co_filename, code.co_name, code.co_firstlineno,
                           code.co_lnotab, code.co_freevars, code.co_cellvars)
    else:
        return types.CodeType(code.co_argcount, code.co_nlocals, code.co_stacksize, code.co_flags,
                              co_code, code.co_consts, code.co_names, code.co_varnames,
                              code.co_filename, code.co_name, code.co_firstlineno,
                              code.co_lnotab, code.co_freevars, code.co_cellvars)

try:
    a = bytearray()
except:
    class bytearray(object):
        def __init__(self, s):
            self.l = map(ord, s)
        def __setitem__(self, idx, val):
            self.l[idx] = val
        def __getitem__(self, idx):
            return self.l[idx]
        def __str__(self):
            return ''.join(map(chr, self.l))
        def __len__(self):
            return len(self.l)

# Automatically generated opcode substitution table for v1
# A different dropbox table for v.2 table is at
#   https://github.com/kholia/dedrop/blob/master/src/dedrop-v2/dedrop-v2.py
# They are similar but not the same.
table = {  0:   0,   1:  87,  2:  66,  4:  25,  6:  55,  7:  62,
           9:  71,  10:  79, 12:  21, 13:   4, 14:  72, 15:   1, 16: 30,
          17:  31,  18:  32, 19:  33, 22:  63,
          26:  86,  29:  56, 31:  60,
          33:  73,  34:  15, 35:  74, 36:  20, 38:  12, 39: 68, 40:  80,
          41:  22,  42:  89, 43:  26, 50:  64, 51:  82, 52:  23, 54: 11,
          55:  24,
          56:  84,  59:   2, 60:   3, 61:  40, 62:  41, 63:  42, 64: 43,
          65:  85,  66:  83, 67:  88, 68:  18, 69:  61, 70: 116, 71: 126,
          72: 100,
          73: 110,  74: 120, 75: 122, 76: 132,
          77: 133,  78: 104, 79: 101, 80: 102,
          81:  93,  82: 125, 83: 111, 84:  95, 85: 134, 86: 105,
          88: 107,  89: 108, 90: 112, 91: 130, 92: 124,
          93:  92,  94:  91, 95:  90,
          97: 135,  99: 136, 100: 137, 101: 106,
         102: 131, 103: 113, 104:  99,
         105:  97, 106: 121, 107: 103, 111: 140,
         112: 141,
         113: 142}

# manual mapping of the rest
table[37] = 81
table[28] = 19
table[87] = 96
table[21] = 65
table[96] = 119
table[8] = 57
table[32] = 28
table[44] = 50
table[45] = 51
table[46] = 52
table[47] = 53
table[23] = 78
table[24] = 77
table[3] = 59
table[11] = 75
table[58] = 76

misses = {}

def patch(code):
    code = bytearray(code)
    i = 0
    n = len(code)
    while i < n:
        op = code[i]
        if not op in table:
            print("missing opcode %d. code: " % op, repr(str(code)))
            misses[op] = misses.get(op, 0) + 1
        code[i] = table.get(op,op)
        i += 1
        if table.get(op,op) >= 90: # opcode.HAVE_ARGUMENT:
            i += 2
    return bytes(code) if PYTHON3 else str(code)

try: from __pypy__ import builtinify
except ImportError: builtinify = lambda f: f

@builtinify
def loads(s):
    """
    xdis.marshal.load() but with its dispatch load_code() function replaced
    with our decoding version.
    """
    um = xmarshal._FastUnmarshaller(s)
    um.dispatch[xmarshal.TYPE_CODE] = load_code
    return um.load()

def fix_dropbox_pyc(fp, fixed_pyc='/tmp/test.pyc'):
    source_size = struct.unpack("I", fp.read(4))[0] # size mod 2**32
    ts = fp.read(4)
    timestamp = struct.unpack("I", ts)[0]
    b = fp.read()
    co = loads(b)
    return 2.5, timestamp, 62131, co, False, source_size

def fix_dir(path):
    import os
    for root, dirs, files in os.walk(path):
        for name in files:
            if not name.endswith('pyc'): continue
            name = os.path.join(root, name)
            print("fixing", name)
            data = open(name).read()
            try:
                c = xmarshal.loads(data[8:])
            except Exception as e:
                print("error", e, repr(e))
                # print repr(data[8:])
                continue
            # fix the version indicator and save
            open(name, "w").write(
                "\xb3\xf2\r\n" + data[4:8] + xmarshal.dumps(c))

if __name__ == '__main__':
    import os, sys
    if sys.argv != 2:
        print("Usage: %s python-file" % os.path.basename(sys.argv[0]))
        sys.exit(1)

    fix_dir(sys.argv[1])

# for the sake of completeness, here are the code-fragments to automatically generate
# the opcode substitution table
if 0:
    import marshal
    def fill(c, d):
        if len(c.co_code) != len(d.co_code):
            print("len mismatch", c, d)
            return
        for i, j in zip(c.co_code, d.co_code):
            # if i in m and not table[i] == j:
            #    print "mismatch %c (%x) => %c (%x)" % (ord(i),ord(i),ord(j),ord(j))
            v = table.setdefault(i, {})
            v[j] = v.get(j, 0) + 1
            pass
        return

    for root, dirs, files in os.walk('library'):
        for name in files:
            name = os.path.join(root, name)
            if not name.endswith('pyc'): continue
            f2 = os.path.join('/tmp/python-defaults-2.7.2/Python-2.5.4/Lib', name[8:])
            if not os.path.exists(f2): continue
            print("loading", name)
            try:
                c = xmarshal.loads(open(name).read()[8:])
            except:
                print("error", name)
                continue
            d =  marshal.loads(open(f2).read()[8:])
            fill(c, d)
            codes_c = filter(lambda x: type(x) == type(c), c.co_consts)
            codes_d = filter(lambda x: type(x) == type(c), d.co_consts)
            for i, j in zip(codes_c, codes_d):
                fill(i, j)
                pass
            pass

    def print_table(m):
        k = m.keys()
        k.sort()
        table = {}
        for i in k:
            print("%c (%02x %s) =>" %
                  (ord(i), ord(i), bin(ord(i))), end='')
            for j, count in m[i].iteritems():
                if j == i: continue
                table[ord(i)] = ord(j)
                print("\t%c (%02x %s) [%d]" %
                      (ord(j), ord(j), bin(ord(j)), count), end="")
                # print("%c (%02x %s) => %c (%02x %s)\t%d\t%s" % (ord(i),ord(i),bin(ord(i)),ord(j),ord(j),bin(ord(j)),ord(j)-ord(i),bin(ord(i)^ord(j)|0x100).replace('0', ' ')))
            print()
        return table

    import re
    re.compile('offset loc_(\w+)').findall('dd offset loc_8096DC4, offset loc_8096963, offset loc_8095462')

    def load(name):
        a = re.compile('offset loc_(\w+)').findall(open(name).read())
        a = [int(i, 16) for i in a]
        c = a[:]
        c.sort()
        c = [(i, c.index(i)) for i in a]
        d = {}
        for i, (addr, pos) in enumerate(c):
            d[addr] = (i, pos)
        return c, d



# (C) Copyright 2017 by Rocky Bernstein
"""
CPython 2.7 bytecode opcodes

This is a like Python 2.7's opcode.py with some classification
of stack usage.
"""

from xdis.opcodes.base import (
    def_op, compare_op, finalize_opcodes, format_extended_arg,
    init_opdata, jabs_op, jrel_op, name_op, rm_op, update_pj3)
import xdis.opcodes.opcode_26 as opcode_26

version = 2.7

l = locals()
init_opdata(l, opcode_26, version)

# Below are opcode changes since Python 2.6
rm_op(l, 'BUILD_MAP',     104)
rm_op(l, 'LOAD_ATTR',     105)
rm_op(l, 'COMPARE_OP',    106)
rm_op(l, 'IMPORT_NAME',   107)
rm_op(l, 'IMPORT_FROM',   108)
rm_op(l, 'JUMP_IF_FALSE', 111)
rm_op(l, 'EXTENDED_ARG',  143)
rm_op(l, 'JUMP_IF_TRUE',  112)

def_op(l, 'LIST_APPEND',            94, 2, 1) # Calls list.append(TOS[-i], TOS).
                                              # Used to implement list comprehensions.
def_op(l, 'BUILD_SET',             104)       # Number of set items
def_op(l, 'BUILD_MAP',             105)
name_op(l, 'LOAD_ATTR',            106)
compare_op(l, 'COMPARE_OP',        107)

name_op(l, 'IMPORT_NAME',          108,  2,  1)  # Index in name list
name_op(l, 'IMPORT_FROM',          109,  0,  1)

jabs_op(l, 'JUMP_IF_FALSE_OR_POP', 111) # Target byte offset from beginning of code
jabs_op(l, 'JUMP_IF_TRUE_OR_POP',  112)  # ""
jabs_op(l, 'POP_JUMP_IF_FALSE',    114)  # ""
jabs_op(l, 'POP_JUMP_IF_TRUE',     115)  # ""
jrel_op(l, 'SETUP_WITH',           143,  0,  2)

def_op(l, 'EXTENDED_ARG',          145)
def_op(l, 'SET_ADD',               146,  1,  0)  # Calls set.add(TOS1[-i], TOS).
                                                 # Used to implement set comprehensions.
def_op(l, 'MAP_ADD',               147,  2,  1)  # Calls dict.setitem(TOS1[-i], TOS, TOS1)
                                                 # Used to implement dict comprehensions.

update_pj3(globals(), l)

opcode_arg_fmt = {
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)



"""Python disassembly functions specific to wordcode from Python 3.6+
"""
from xdis import PYTHON3
from xdis.bytecode import op_has_argument

def unpack_opargs_wordcode(code, opc):
    extended_arg = 0
    try:
        n = len(code)
    except TypeError:
        code = code.co_code
        n = len(code)

    for i in range(0, n, 2):
        op = code[i]
        if op_has_argument(op, opc):
            if isinstance(code[i+1], str):
                arg = ord(code[i+1]) | extended_arg
            else:
                arg = code[i+1] | extended_arg
            extended_arg = (arg << 8) if op == opc.EXTENDED_ARG else 0
        else:
            arg = None
        yield (i, op, arg)

def findlinestarts(code, dup_lines=False):
    """Find the offsets in a byte code which are start of lines in the source.

    Generate pairs (offset, lineno) as described in Python/compile.c.

    """
    if PYTHON3:
        byte_increments = code.co_lnotab[0::2]
        line_increments = code.co_lnotab[1::2]
    else:
        byte_increments = [ord(c) for c in code.co_lnotab[0::2]]
        line_increments = [ord(c) for c in code.co_lnotab[1::2]]

    lastlineno = None
    lineno = code.co_firstlineno
    addr = 0
    for byte_incr, line_incr in zip(byte_increments, line_increments):
        if byte_incr:
            if (lineno != lastlineno or
                (not dup_lines and 0 < byte_incr < 255)):
                yield (addr, lineno)
                lastlineno = lineno
            addr += byte_incr
        if line_incr >= 0x80:
            # line_increments is an array of 8-bit signed integers
            line_incr -= 0x100
        lineno += line_incr
    if (lineno != lastlineno or
        (not dup_lines and 0 < byte_incr < 255)):
        yield (addr, lineno)


def get_jump_targets(code, opc):
    """Returns a list of instruction offsets in the supplied bytecode
    which are the targets of some sort of jump instruction.
    """
    offsets = []
    for offset, op, arg in unpack_opargs_wordcode(code, opc):
        if arg is not None:
            jump_offset = -1
            if op in opc.JREL_OPS:
                jump_offset = offset + 2 + arg
            elif op in opc.JABS_OPS:
                jump_offset = arg
            if jump_offset >= 0:
                if jump_offset not in offsets:
                    offsets.append(jump_offset)
    return offsets

def get_jump_target_maps(code, opc):
    """Returns a dictionary where the key is an offset and the values are
    a list of instruction offsets which can get run before that
    instruction. This includes jump instructions as well as non-jump
    instructions. Therefore, the keys of the dictionary are reachible
    instructions. The values of the dictionary may be useful in control-flow
    analysis.
    """
    offset2prev = {}
    prev_offset = -1
    for offset, op, arg in unpack_opargs_wordcode(code, opc):
        if prev_offset >= 0:
            prev_list = offset2prev.get(offset, [])
            prev_list.append(prev_offset)
            offset2prev[offset] = prev_list
        prev_offset = offset
        if op in opc.NOFOLLOW:
            prev_offset = -1
        if arg is not None:
            jump_offset = -1
            if op in opc.JREL_OPS:
                jump_offset = offset + 2 + arg
            elif op in opc.JABS_OPS:
                jump_offset = arg
            if jump_offset >= 0:
                prev_list = offset2prev.get(jump_offset, [])
                prev_list.append(offset)
                offset2prev[jump_offset] = prev_list
    return offset2prev


findlabels = get_jump_targets



# (C) Copyright 2016-2017 by Rocky Bernstein
"""
CPython 3.7 bytecode opcodes

This is a like Python 3.7's opcode.py with some classification
of stack usage.
"""

from xdis.opcodes.base import(
    def_op, finalize_opcodes,
    format_extended_arg36,
    init_opdata, name_op,
    update_pj3
    )

import xdis.opcodes.opcode_36 as opcode_36

version = 3.7

l = locals()

init_opdata(l, opcode_36, version)

### Opcodes that have changed drastically ####


# BUILD_MAP_UNPACK_WITH_CALL oparg

# oparg is the number of unpacked mappings which no longer limited by
# 255. As in BUILD_MAP_UNPACK.The location of the function is `oparg +
# 2`.


# CALL_FUNCTION oparg

# oparg is the number of positional arguments on the stack which no
# longer limited by 255.


# CALL_FUNCTION_KW oparg

# This is a different opcode from before, with the name of a similar
# opcode as before. It s like CALL_FUNCTION but values of keyword
# arguments are pushed on the stack after values of positional
# arguments, and then a constant tuple of keyword names is pushed on
# the top of the stack.  *oparg* is the sum of numbers of positional
# and keyword arguments.  The number of keyword arguments is
# determined by the length of the tuple of keyword names.

# CALL_FUNCTION_EX** takes 2 to 3 arguments on the stack: the
# function, the tuple of positional arguments, and optionally the dict
# of keyword arguments if bit 0 of *oparg* is 1.


# MAKE_FUNCTION oparg

# This is a different opcode from before.

# The tuple of default values for positional-or-keyword parameters,
# the dict of default values for keyword-only parameters, the dict of
# annotations and the closure are pushed on the stack if corresponding
# bit (0-3) is set. They are followed by the code object and the
# qualified name of the function.


# These are new since Python 3.7
name_op(l, 'LOAD_METHOD', 160)
def_op(l, 'CALL_METHOD', 161)

MAKE_FUNCTION_FLAGS = tuple("default keyword-only annotation closure".split())

def format_MAKE_FUNCTION_arg(flags):
    pattr = ''
    for flag in MAKE_FUNCTION_FLAGS:
        bit = flags & 1
        if bit:
            if pattr:
                pattr += ", " + flag
            else:
                pattr = flag
                pass
            pass
        flags >>= 1
    return pattr

def format_value_flags(flags):
    if (flags & 0x03) == 0x00:
        return ''
    elif (flags & 0x03) == 0x01:
        return '!s'
    elif (flags & 0x03) == 0x02:
        return '!r'
    elif (flags & 0x03) == 0x03:
        return '!a'
    elif (flags & 0x04) == 0x04:
        # pop fmt_spec from the stack and use it, else use an
        # empty fmt_spec.
        return ''

opcode_arg_fmt = {
    'MAKE_FUNCTION': format_MAKE_FUNCTION_arg,
    'FORMAT_VALUE': format_value_flags,
    'EXTENDED_ARG': format_extended_arg36
}

update_pj3(globals(), l)

finalize_opcodes(l)



# (C) Copyright 2017 by Rocky Bernstein
"""
CPython 3.1 bytecode opcodes

This is a like Python 3.1's opcode.py with some classification
of stack usage.
"""

from xdis.opcodes.base import (
    def_op, format_extended_arg, finalize_opcodes, init_opdata,
    rm_op, update_pj3)

from xdis.opcodes.opcode_3x import format_MAKE_FUNCTION_arg

import xdis.opcodes.opcode_32 as opcode_32

l = locals()

version = 3.1

init_opdata(l, opcode_32, version)

# These are in Python 3.2 but not in Python 3.1
rm_op(l, 'DUP_TOP_TWO',    5)
rm_op(l, 'DELETE_DEREF', 138)
rm_op(l, 'SETUP_WITH',   143)

# These are in Python 3.1 but not Python 3.2
def_op(l, 'ROT_FOUR', 5)
def_op(l, 'DUP_TOPX', 99)

# This op is in 3.2 but its opcode is a 144 instead
def_op(l, 'EXTENDED_ARG', 143)

update_pj3(globals(), l)

opcode_arg_fmt = {
    'MAKE_FUNCTION': format_MAKE_FUNCTION_arg,
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)



# (C) Copyright 2017 by Rocky Bernstein
"""
CPython 2.5 bytecode opcodes

This is a like Python 2.5's opcode.py with some classification
of stack usage.
"""

import xdis.opcodes.opcode_24 as opcode_24
from xdis.opcodes.base import (
    def_op, init_opdata,
    finalize_opcodes, format_extended_arg,
    update_pj2)

version = 2.5

l = locals()
init_opdata(l, opcode_24, version)

# Bytecodes added in 2.5 from 2.4
def_op(l, 'WITH_CLEANUP', 81)

# FIXME remove (fix uncompyle6)
update_pj2(globals(), l)

opcode_arg_fmt = {
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)



# (C) Copyright 2017 by Rocky Bernstein
"""
PYPY 3.2 opcodes

This is a like Python 3.2's opcode.py with some classification
of stack usage.
"""

from xdis.opcodes.base import (
    finalize_opcodes, init_opdata, jrel_op, name_op,
    varargs_op, update_pj3)

version = 3.2

import xdis.opcodes.opcode_32 as opcode_32

l = locals()
init_opdata(l, opcode_32, version, is_pypy=True)

## FIXME: DRY common PYPY opcode additions

# PyPy only
# ----------
name_op(l, 'LOOKUP_METHOD',  201,  1, 2)
varargs_op(l, 'CALL_METHOD', 202, -1, 1)
l['hasvargs'].append(202)

# Used only in single-mode compilation list-comprehension generators
varargs_op(l, 'BUILD_LIST_FROM_ARG', 203)

# Used only in assert statements
jrel_op(l, 'JUMP_IF_NOT_DEBUG',      204, conditional=True)

# There are no opcodes to remove or change.
# If there were, they'd be listed below.

# FIXME remove (fix uncompyle6)
update_pj3(globals(), l)

finalize_opcodes(l)



"""Facilitates importing opmaps for the a given Python version"""
import sys
from xdis import IS_PYPY
from xdis.magics import canonic_python_version

from xdis.opcodes import opcode_15 as opcode_15
from xdis.opcodes import opcode_20 as opcode_20
from xdis.opcodes import opcode_21 as opcode_21
from xdis.opcodes import opcode_22 as opcode_22
from xdis.opcodes import opcode_23 as opcode_23
from xdis.opcodes import opcode_24 as opcode_24
from xdis.opcodes import opcode_25 as opcode_25
from xdis.opcodes import opcode_26 as opcode_26
from xdis.opcodes import opcode_27 as opcode_27
from xdis.opcodes import opcode_30 as opcode_30
from xdis.opcodes import opcode_31 as opcode_31
from xdis.opcodes import opcode_32 as opcode_32
from xdis.opcodes import opcode_33 as opcode_33
from xdis.opcodes import opcode_34 as opcode_34
from xdis.opcodes import opcode_35 as opcode_35
from xdis.opcodes import opcode_36 as opcode_36
from xdis.opcodes import opcode_37 as opcode_37

from xdis.opcodes import opcode_26pypy as opcode_26pypy
from xdis.opcodes import opcode_27pypy as opcode_27pypy
from xdis.opcodes import opcode_32pypy as opcode_32pypy
from xdis.opcodes import opcode_35pypy as opcode_35pypy

op_imports = {
    '1.5'  : opcode_15,
    1.5    : opcode_15,
    '2.0'  : opcode_20,
    2.0    : opcode_20,
    '2.1'  : opcode_21,
    2.1    : opcode_21,
    '2.2'  : opcode_22,
    2.2    : opcode_22,
    '2.3a0': opcode_23,
    2.3    : opcode_23,
    '2.4b1': opcode_24,
    2.4    : opcode_24,
    '2.5c2': opcode_25,
    2.5:     opcode_25,
    '2.6a1': opcode_26,
    2.6    : opcode_26,
    '2.7'  : opcode_27,
    2.7    : opcode_27,
    '3.0'  : opcode_30,
    3.0    : opcode_30,
    '3.1'  : opcode_31,
    3.1    : opcode_31,
    '3.2'  : opcode_32,
    '3.2a2': opcode_32,
    3.2    : opcode_32,
    '3.3a4': opcode_33,
    3.3    : opcode_33,
    '3.4'  : opcode_34,
    '3.4rc2'  : opcode_34,
    3.4    : opcode_34,
    '3.5'  : opcode_35,
    '3.5.3': opcode_35,  #FIXME:  Not sure if this is right
    '3.5.4': opcode_35,  # not sure if this is right
    3.5    : opcode_35,
    '3.6.0rc1': opcode_36,
    3.6    : opcode_36,
    '3.7.0alpha0': opcode_37,
    3.7    : opcode_37,

    '2.6pypy':  opcode_26pypy,
    '2.7pypy':  opcode_27pypy,
    '3.2pypy':  opcode_32pypy,
    '3.5pypy':  opcode_35pypy,
    }

for k, v in canonic_python_version.items():
    if v in op_imports:
        op_imports[k] = op_imports[v]

def get_opcode_module(version_info=None, variant=None):
    # FIXME: DRY with magics.sysinfo2float()
    if version_info is None:
        version_info = sys.version_info
        if variant is None and IS_PYPY:
            variant = 'pypy'

    vers_str = '.'.join([str(v) for v in version_info[0:3]])
    if version_info[3] != 'final':
        vers_str += ''.join([str(v) for v in version_info[3:]])
    if variant is None:
        try:
            import platform
            variant = platform.python_implementation()
            if platform in ('Jython', 'Pyston'):
                vers_str += variant
                pass
        except ImportError:
            # Python may be too old, e.g. < 2.6 or implementation may
            # just not have platform
            pass
    else:
        vers_str += variant
    return op_imports[canonic_python_version[vers_str]]


if __name__ == '__main__':
    print(get_opcode_module())



# (C) Copyright 2017 by Rocky Bernstein
"""
CPython 1.5 bytecode opcodes

This is used in bytecode disassembly. This is similar to the
opcodes in Python's dis.py library.
"""

# These are used from outside this module
from xdis.bytecode import findlabels, findlinestarts
from xdis.opcodes.base import (
    compare_op, const_op,
    def_op, jabs_op, jrel_op,
    finalize_opcodes, format_extended_arg,
    local_op, name_op, nargs_op,
    # Although these aren't used here, they are exported
    varargs_op, update_pj2
    )

version = 1.5

cmp_op = ('<', '<=', '==', '!=', '>', '>=', 'in', 'not in', 'is',
        'is not', 'exception match', 'BAD')

# Opcodes greater than 90 take an instruction operand or "argument"
# as opcode.py likes to call it.
HAVE_ARGUMENT = 90

l = locals()
l['python_version'] = version
l['cmp_op'] = cmp_op
l['HAVE_ARGUMENT'] = HAVE_ARGUMENT

# These are just to silence the import above
l['findlindstarts'] = findlinestarts
l['findlabels'] = findlabels

# FIXME: can we DRY this even more?

hascompare   = []
hascondition = [] # conditional operator; has jump offset
hasconst     = []
hasfree      = []
hasjabs      = []
hasjrel      = []
haslocal     = []
hasname      = []
hasnargs     = []  # For function-like calls
hasvargs     = []  # Similar but for operators BUILD_xxx
nofollow     = []  # Instruction doesn't fall to the next opcode

# oppush[op] => number of stack entries pushed
# -9 means handle special. Note his forces oppush[i] - oppop[i] negative
oppush = [0] * 256

# oppop[op] => number of stack entries popped
# -1 means handle special.
oppop  = [0] * 256

opmap = {}
opname = [''] * 256
for op in range(256): opname[op] = '<%r>' % (op,)
del op

# Instruction opcodes for compiled code
# Blank lines correspond to available opcodes

def_op(l, 'STOP_CODE',       0,  0,  0, fallthrough=False)
def_op(l, 'POP_TOP', 1)
def_op(l, 'ROT_TWO', 2)
def_op(l, 'ROT_THREE', 3)
def_op(l, 'DUP_TOP', 4)

def_op(l, 'UNARY_POSITIVE',  10, 1, 1)
def_op(l, 'UNARY_NEGATIVE',  11, 1, 1)
def_op(l, 'UNARY_NOT',       12, 1, 1)
def_op(l, 'UNARY_CONVERT',   13, 1, 1)

def_op(l, 'UNARY_INVERT',    15, 1, 1)

def_op(l, 'BINARY_POWER',    19, 1, 1)

def_op(l, 'BINARY_MULTIPLY', 20, 2, 1)
def_op(l, 'BINARY_DIVIDE',   21, 2, 1)
def_op(l, 'BINARY_MODULO',   22, 2, 1)
def_op(l, 'BINARY_ADD',      23, 2, 1)
def_op(l, 'BINARY_SUBTRACT', 24, 2, 1)
def_op(l, 'BINARY_SUBSCR',   25, 2, 1)

def_op(l, 'SLICE+0',         30, 1, 1)
def_op(l, 'SLICE+1',         31, 2, 1)
def_op(l, 'SLICE+2',         32, 2, 1)
def_op(l, 'SLICE+3',         33, 3, 1)

def_op(l, 'STORE_SLICE+0',   40, 2, 0)
def_op(l, 'STORE_SLICE+1',   41, 3, 0)
def_op(l, 'STORE_SLICE+2',   42, 3, 0)
def_op(l, 'STORE_SLICE+3',   43, 4, 0)

def_op(l, 'DELETE_SLICE+0',  50, 1, 0)
def_op(l, 'DELETE_SLICE+1',  51, 2, 0)
def_op(l, 'DELETE_SLICE+2',  52, 2, 0)
def_op(l, 'DELETE_SLICE+3',  53, 3, 0)

def_op(l, 'STORE_SUBSCR',    60, 3, 0) # Implements TOS1[TOS] = TOS2.
def_op(l, 'DELETE_SUBSCR',   61, 2, 0) # Implements del TOS1[TOS].

def_op(l, 'BINARY_LSHIFT',   62, 2, 1)
def_op(l, 'BINARY_RSHIFT',   63, 2, 1)
def_op(l, 'BINARY_AND',      64, 2, 1)
def_op(l, 'BINARY_XOR',      65, 2, 1)
def_op(l, 'BINARY_OR',       66, 2, 1)

def_op(l, 'PRINT_EXPR',      70, 1, 0)
def_op(l, 'PRINT_ITEM',      71, 1, 0)
def_op(l, 'PRINT_NEWLINE',   72, 1, 0)

def_op(l, 'BREAK_LOOP',      80, 0, 0)

def_op(l, 'LOAD_LOCALS',     82, 0, 1)
def_op(l, 'RETURN_VALUE',    83,  1,  0, fallthrough=False)

def_op(l, 'EXEC_STMT',       85, 3, 0)

def_op(l, 'POP_BLOCK',       87, 0, 0)
def_op(l, 'END_FINALLY',     88, 1, 0)
def_op(l, 'BUILD_CLASS',     89, 3, 0)

# HAVE_ARGUMENT = 90               # Opcodes from here have an argument:

name_op(l, 'STORE_NAME',           90,  1,  0)  # Operand is in name list
name_op(l, 'DELETE_NAME',          91,  0,  0)  # ""
varargs_op(l, 'UNPACK_TUPLE', 92)               # Number of tuple items
def_op(l, 'UNPACK_LIST', 93)	                # Number of list items
name_op(l, 'STORE_ATTR',           95,  2,  0)  # Operand is in name list
name_op(l, 'DELETE_ATTR',          96,  1,  0)  # ""
name_op(l, 'STORE_GLOBAL',         97,  1,  0)  # ""
name_op(l, 'DELETE_GLOBAL',        98,  0,  0)  # ""

const_op(l, 'LOAD_CONST',         100,  0,  1)  # Operand is in const list
name_op(l, 'LOAD_NAME',           101,  0,  1)  # Operand is in name list
varargs_op(l, 'BUILD_TUPLE',      102, -1,  1)  # Number of tuple items
varargs_op(l, 'BUILD_LIST',       103, -1,  1)  # Number of list items
varargs_op(l, 'BUILD_MAP',        104, -1,  1)  # Always zero for now
name_op(l, 'LOAD_ATTR',           105,  1,  1)  # Operand is in name list
compare_op(l, 'COMPARE_OP',       106,  2,  1)  # Comparison operator

name_op(l, 'IMPORT_NAME',         107,  2,  1) # Operand is in name list
name_op(l, 'IMPORT_FROM',         108,  0,  1) # Operand is in name list

jrel_op(l, 'JUMP_FORWARD',        110,  0,  0)  # Number of bytes to skip
jrel_op(l, 'JUMP_IF_FALSE',       111,  1,  1, True)  # ""
jrel_op(l, 'JUMP_IF_TRUE',        112,  1,  1, True)  # ""
jabs_op(l, 'JUMP_ABSOLUTE',       113,  0,  0)  # Target byte offset from beginning of code
def_op(l, 'FOR_LOOP',             114)	        # Number of bytes to skip

name_op(l, 'LOAD_GLOBAL',         116,  0,  1)  # Operand is in name list

jrel_op(l, 'SETUP_LOOP',          120,  0,  0, conditional=True)  # Distance to target address
jrel_op(l, 'SETUP_EXCEPT',        121,  0,  0)  # ""
jrel_op(l, 'SETUP_FINALLY',       122,  0,  0)  # ""

local_op(l, 'LOAD_FAST',          124,  0,  1)  # Local variable number
local_op(l, 'STORE_FAST',         125,  1,  0)  # Local variable number
local_op(l, 'DELETE_FAST',        126)          # Local variable number

def_op(l, 'SET_LINENO', 127)	   # Current line number

def_op(l, 'RAISE_VARARGS',        130, -1,  0, fallthrough=False)
                                                # Number of raise arguments (1, 2, or 3)
nargs_op(l, 'CALL_FUNCTION',      131, -1,  1)  # #args + (#kwargs << 8)

def_op(l, 'MAKE_FUNCTION',        132, -1,  1)  # Number of args with default values
varargs_op(l, 'BUILD_SLICE',      133, -1,  1)  # Number of items

def_op(l, 'EXTENDED_ARG', 143)
EXTENDED_ARG = 143

fields2copy = """cmp_op hasjabs""".split()

update_pj2(globals(), l)

opcode_arg_fmt = {
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)



# Python 3.6 format strings.
# Example taken right out of PEP 484
import datetime
name = 'Fred'
age = 50
anniversary = datetime.date(1991, 10, 12)
print(f'My name is {name}, my age next year is {age+1}, my anniversary is {anniversary:%A, %B %d, %Y}.')
print(f'He said his name is {name!r}.')



# (C) Copyright 2017 by Rocky Bernstein
"""
PYPY 2.6 opcodes

This is a like Python 2.6's opcode.py with some classification
of stack usage.
"""

import xdis.opcodes.opcode_26 as opcode_26
from xdis.opcodes.base import (
    finalize_opcodes, init_opdata,
    jrel_op, name_op, varargs_op,
    update_pj2
    )

version = 2.6

l = locals()
init_opdata(l, opcode_26, version, is_pypy=True)

# FIXME: DRY common PYPY opcode additions

# PyPy only
# ----------
name_op(l, 'LOOKUP_METHOD',   201,  1, 2)
varargs_op(l,  'CALL_METHOD', 202, -1, 1)
l['hasnargs'].append(202)

# Used only in single-mode compilation list-comprehension generators
varargs_op(l, 'BUILD_LIST_FROM_ARG', 203)

# Used only in assert statements
jrel_op(l, 'JUMP_IF_NOT_DEBUG',      204, conditional=True)

# FIXME remove (fix uncompyle6)
update_pj2(globals(), l)

finalize_opcodes(l)



"""
Everything you ever wanted to know about Python versions and their
magic numbers. And a little bit more...
"""
import imp, re, struct, sys
from xdis import IS_PYPY

def int2magic(magic):
    if (sys.version_info >= (3, 0)):
        return struct.pack('Hcc', magic, bytes('\r', 'utf-8'), bytes('\n', 'utf-8'))
    else:
        return struct.pack('Hcc', magic, '\r', '\n')

def magic2int(magic):
    return struct.unpack('Hcc', magic)[0]

# The magic integer for the current running Python interpreter
PYTHON_MAGIC_INT = magic2int(imp.get_magic())

by_magic = {}
by_version = {}

def __by_version(magics):
    for m, v in list(magics.items()):
        if m not in by_magic:
            by_magic[m] = set([v])
        else:
            by_magic[m].add(v)
        by_version[v] = m
    return by_version

versions = {
    # Magic word to reject .pyc files generated by other Python versions.
    # It should change for each incompatible change to the bytecode.
    #
    # The value of CR and LF is incorporated so if you ever read or write
    # a .pyc file in text mode the magic number will be wrong; also, the
    # Apple MPW compiler swaps their values, botching string constants.
    #
    # The magic numbers must be spaced apart at least 2 values, as the
    # -U interpeter flag will cause MAGIC+1 being used. They have been
    # odd numbers for some time now.
    #
    # There were a variety of old schemes for setting the magic number.
    # The current working scheme is to increment the previous value by
    # 10.
    #
    # Starting with the adoption of PEP 3147 in Python 3.2, every bump in magic
    # number also includes a new "magic tag", i.e. a human readable string used
    # to represent the magic number in __pycache__ directories.  When you change
    # the magic number, you must also set a new unique magic tag.  Generally this
    # can be named after the Python major version of the magic number bump, but
    # it can really be anything, as long as it's different than anything else
    # that's come before.  The tags are included in the following table, starting
    # with Python 3.2a0.

    # taken from from Python/import.c, importlib/_bootstrap.py and other sources
    # magic,          sample version number
    int2magic(11913): '1.3', #
    int2magic(5892):  '1.4', #
    int2magic(20121): '1.5', # 1.5, 1.5.1, 1.5.2
    int2magic(50428): '1.6', # 1.6
    int2magic(50823): '2.0', # 2.0, 2.0.1
    int2magic(60202): '2.1', # 2.1, 2.1.1, 2.1.2
    int2magic(60717): '2.2', # 2.2
    int2magic(62011): '2.3a0',
    int2magic(62021): '2.3a0', # ! Two magics one version
    int2magic(62041): '2.4a0',
    int2magic(62051): '2.4a3',
    int2magic(62061): '2.4b1',
    int2magic(62071): '2.5a0',
    int2magic(62081): '2.5a0', # ast-branch
    int2magic(62091): '2.5a0', # with
    int2magic(62092): '2.5a0', # changed WITH_CLEANUP opcode
    int2magic(62101): '2.5b3', # fix wrong code: for x, in ...
    int2magic(62111): '2.5b3', # fix wrong code: x += yield
    int2magic(62121): '2.5c1', # fix wrong lnotab with for loops and
                               #  storing constants that should have been removed
    int2magic(62131): '2.5c2', # fix wrong code: for x, in ... in
                               # listcomp/genexp)
    int2magic(62135): '2.5dropbox', # Dropbox-modified Python 2.5
                                    # used in versions 1.1x and before of Dropbox
    int2magic(62151): '2.6a0',   # peephole optimizations & STORE_MAP
    int2magic(62161): '2.6a1',   # WITH_CLEANUP optimization
    int2magic(62171): '2.7a0',   # optimize list comprehensions/change
                                 # LIST_APPEND
    int2magic(62181): '2.7a0+1', # optimize conditional branches:
                                 #  introduce POP_JUMP_IF_FALSE and
                                 # POP_JUMP_IF_TRUE
    int2magic(62191): '2.7a0+2', # introduce SETUP_WITH
    int2magic(62201): '2.7a0+3', # introduce BUILD_SET
    int2magic(62211): '2.7',     # introduce MAP_ADD and SET_ADD

    int2magic(62215): '2.7dropbox', # Dropbox-modified Python 2.7
                                    # used in versions 1.2-1.6 or so of
                                    # Dropbox

    int2magic(62211+7): '2.7pypy', # PyPy including pypy-2.6.1, pypy-5.0.1
                                   # PyPy adds 7 to the corresponding CPython nmber
    int2magic(3000): '3.000',
    int2magic(3010): '3.000+1',  # removed UNARY_CONVERT
    int2magic(3020): '3.000+2',  # added BUILD_SET
    int2magic(3030): '3.000+3',  # added keyword-only parameters
    int2magic(3040): '3.000+4',  # added signature annotations
    int2magic(3050): '3.000+5',  # print becomes a function
    int2magic(3060): '3.000+6',  # PEP 3115 metaclass syntax
    int2magic(3061): '3.000+7',  # string literals become unicode
    int2magic(3071): '3.000+8',  # PEP 3109 raise changes
    int2magic(3081): '3.000+9',  # PEP 3137 make __file__ and __name__ unicode
    int2magic(3091): '3.000+10', # kill str8 interning
    int2magic(3101): '3.000+11', # merge from 2.6a0, see 62151
    int2magic(3103): '3.000+12', # __file__ points to source file
    int2magic(3111): '3.0a4',  # WITH_CLEANUP optimization
    int2magic(3131): '3.0a5',  # lexical exception stacking, including POP_EXCEPT)
    int2magic(3141): '3.1a0',  # optimize list, set and dict comprehensions
    int2magic(3151): '3.1a0+', # optimize conditional branches
    int2magic(3160): '3.2a0',  # add SETUP_WITH
    int2magic(3170): '3.2a1',  # add DUP_TOP_TWO, remove DUP_TOPX and ROT_FOUR
    int2magic(3180): '3.2a2',  # 3.2a2 (add DELETE_DEREF)
    int2magic(3180+7): '3.2pypy',  # Python 3.2.5 - PyPy 2.3.4
                                   # PyPy adds 7 to the corresponding CPython number
    int2magic(3190): '3.3a0',  # __class__ super closure changed
    int2magic(3200): '3.3a0+', # __qualname__ added
    int2magic(3220): '3.3a1',  # changed PEP 380 implementation
    int2magic(3210): '3.3a2',  # added size modulo 2**32 to the pyc header
                               # NOTE: 3.3a2 is our name, other places call it 3.3
                               # but most 3.3 versions are 3.3a4 which comes next.
                               # FIXME: figure out what the history is and
                               # what the right thing to do if this isn't it.
    int2magic(3230): '3.3a4',  # revert changes to implicit __class__ closure
    int2magic(3250): '3.4a1',  # evaluate positional default arg
                               # keyword-only defaults)
    int2magic(3260): '3.4a1+1', # add LOAD_CLASSDEREF;
                                # allow locals of class to override free vars
    int2magic(3270): '3.4a1+2', # various tweaks to the __class__ closure
    int2magic(3280): '3.4a1+3',   # remove implicit class argument
    int2magic(3290): '3.4a4',     # changes to __qualname__ computation
    int2magic(3300): '3.4a4+',    # more changes to __qualname__ computation
    int2magic(3310): '3.4rc2',    # alter __qualname__ computation
    int2magic(3350): '3.5',       # 3.5.0, 3.5.1, 3.5.2
    int2magic(3351): '3.5.3',     # 3.5.3
    int2magic(3361): '3.6.0a1',   # 3.6.0a1
    int2magic(3370): '3.6.0a1+1', # 3.6.0a?
    int2magic(3370): '3.6.0a1+2', #
    int2magic(3372): '3.6.0a3',   #
    int2magic(3378): '3.6.0b2',   #
    int2magic(3379): '3.6.0rc1',  #
    int2magic(3390): '3.7.0alpha0',

    # Weird ones
    int2magic(48):    '3.2a2', # WTF? Python 3.2.5 - PyPy 2.3.4
                               # This doesn't follow the rule below
    int2magic(112):   '3.5pypy', # pypy3.5-c-jit-latest
    int2magic(1011):  '2.7.1b3Jython', # jython
    int2magic(22138): '2.7.7Pyston',  # 2.7.8pystem, pyston-0.6.0, pyston-0.6.1
}

magics = __by_version(versions)

# From a Python version given in sys.info, e.g. 3.6.1,
# what is the "canonic" version number, e.g. '3.6.0rc1'
canonic_python_version = {}

def add_canonic_versions(versions, canonic):
    for v in versions.split():
        canonic_python_version[v] = canonic
        magics[v] = magics[canonic]
        try:
            magics[float(v)] = magics[canonic]
        except:
            pass
        pass


    return

add_canonic_versions('1.5.1 1.5.2', '1.5')
add_canonic_versions('2.0.1', '2.0')
add_canonic_versions('2.1.1 2.1.2', '2.1')
add_canonic_versions('2.2.3', '2.2')
add_canonic_versions('2.3 2.3.7', '2.3a0')
add_canonic_versions('2.4 2.4.1 2.4.2 2.4.3 2.4.5 2.4.6', '2.4b1')
add_canonic_versions('2.5 2.5.1 2.5.2 2.5.3 2.5.4 2.5.5 2.5.6', '2.5c2')
add_canonic_versions('2.6 2.6.6 2.6.7 2.6.8 2.6.9', '2.6a1')
add_canonic_versions('2.7.1 2.7.2 2.7.2 2.7.3 2.7.4 2.7.5 2.7.6 2.7.7 '
                     '2.7.8 2.7.9 2.7.10 2.7.11 2.7.12 2.7.13 2.7.14', '2.7')
add_canonic_versions('3.2 3.2.0 3.2.1 3.2.2 3.2.3 3.2.4 3.2.5 3.2.6', '3.2a2')
add_canonic_versions('3.3 3.3.1 3.3.0 3.3.2 3.3.3 3.3.4 3.3.5 3.3.6 3.3.7rc1', '3.3a4')
add_canonic_versions('3.4 3.4.0 3.4.1 3.4.2 3.4.3 3.4.4 3.4.5 3.4.6 3.4.7', '3.4rc2')
add_canonic_versions('3.5.0 3.5.1 3.5.2', '3.5')
add_canonic_versions('3.5.3 3.5.4', '3.5.3')
add_canonic_versions('3.6 3.6.0 3.6.1 3.6.2 3.6.3', '3.6.0rc1')

add_canonic_versions('2.7.10pypy 2.7.13pypy', '2.7pypy')
add_canonic_versions('2.7.3b0Jython', '2.7.1b3Jython')
add_canonic_versions('3.2.5pypy', '3.2pypy')
add_canonic_versions('3.5.3pypy', '3.5pypy')
add_canonic_versions('3.5.3pypy', '3.5pypy')
add_canonic_versions('2.7.8Pyston', '2.7.7Pyston')
add_canonic_versions('3.7 3.7.0', '3.7.0alpha0')

# The canonic version for a canonic version is itself
for v in versions.values():
    canonic_python_version[v] = v

# A set of all Python versions we know about
python_versions = set(canonic_python_version.keys())

def __show(text, magic):
    print(text, struct.unpack('BBBB', magic), struct.unpack('HBB', magic))

def py_str2float(version):
    """Convert a Python version into a 'canonic' floating-point number which
    that can then be used to look up a magic number.
    A runtime error is raised if "version" is not found.
    """
    if version in magics:
        magic = magics[version]
        for v, m in list(magics.items()):
            if m == magic:
                try:
                    return float(canonic_python_version[v])
                except:
                    try:
                        m = re.match('^(\d\.)(\d+)\.(\d+)$', v)
                        if m:
                            return float(m.group(1)+m.group(2)+m.group(3))
                    except:
                        pass
                    pass
                pass
            pass
    raise RuntimeError("Can't find a valid Python version for version %s"
                       % version)
    return

def sysinfo2float(version_info=sys.version_info):
    """Convert a sys.versions_info-compatible list into a 'canonic'
    floating-point number which that can then be used to look up a
    magic number.  Note that this can only be used for released version
    of C Python, not interim development versions, since we can't
    represent that as a floating-point number.

    For handling Pypy, pyston, jython, etc. and interim versions of
    C Python, use sysinfo2magic.
    """
    vers_str = '.'.join([str(v) for v in version_info[0:3]])
    if version_info[3] != 'final':
        vers_str += '.' + ''.join(version_info)
    if IS_PYPY:
        vers_str += 'pypy'
    else:
        try:
            import platform
            platform = platform.python_implementation()
            if platform in ('Jython', 'Pyston'):
                vers_str += platform
                pass
        except ImportError:
            # Python may be too old, e.g. < 2.6 or implementation may
            # just not have platform
            pass
    return py_str2float(vers_str)


def sysinfo2magic(version_info=sys.version_info):
    """Convert a list sys.versions_info compatible list into a 'canonic'
    floating-point number which that can then be used to look up a
    magic number.  Note that this can raise an exception.
    """

    # FIXME: DRY with sysinfo2float()
    vers_str = '.'.join([str(v) for v in version_info[0:3]])
    if version_info[3] != 'final':
        vers_str += ''.join([str(v) for v in version_info[3:]])

    if IS_PYPY:
        vers_str += 'pypy'
    else:
        try:
            import platform
            platform = platform.python_implementation()
            if platform in ('Jython', 'Pyston'):
                vers_str += platform
                pass
        except ImportError:
            # Python may be too old, e.g. < 2.6 or implementation may
            # just not have platform
            pass

    return magics[vers_str]


def test():
    magic_20 = magics['2.0']
    current = imp.get_magic()
    magic_current = by_magic[ current ]
    print(type(magic_20), len(magic_20), repr(magic_20))
    print()
    print('This Python interpreter has version', magic_current)
    print('Magic code: ', PYTHON_MAGIC_INT)
    print(type(magic_20), len(magic_20), repr(magic_20))
    print(sysinfo2float())
    assert sysinfo2magic() == current

if __name__ == '__main__':
    test()



# (C) Copyright 2017 by Rocky Bernstein
"""
CPython 3.4 bytecode opcodes

This is a like Python 3.4's opcode.py with some classification
of stack usage.
"""

from xdis.opcodes.base import (
    def_op, finalize_opcodes,
    format_extended_arg, free_op, init_opdata,
    rm_op, update_pj3)

from xdis.opcodes.opcode_3x import format_MAKE_FUNCTION_arg

import xdis.opcodes.opcode_33 as opcode_33

version = 3.4

l = locals()

init_opdata(l, opcode_33, version)

# These are removed since Python 3.3
rm_op(l, 'STORE_LOCALS', 69)

# These are new since Python 3.3
def_op(l,  'YIELD_FROM',       72)
free_op(l, 'LOAD_CLASSDEREF', 148)

update_pj3(globals(), l)

opcode_arg_fmt = {
    'MAKE_FUNCTION': format_MAKE_FUNCTION_arg,
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)



# (C) Copyright 2017 by Rocky Bernstein
"""
CPython 2.2 bytecode opcodes

This is similar to the opcode portion in Python 2.2's dis.py library.
"""

import xdis.opcodes.opcode_2x as opcode_2x
from xdis.opcodes.base import (
    def_op, init_opdata, finalize_opcodes,
    format_extended_arg, update_pj2)

version = 2.2

l = locals()
init_opdata(l, opcode_2x, version)

# 2.2 Bytecodes not in 2.3
def_op(l, 'FOR_LOOP',   114)
def_op(l, 'SET_LINENO', 127, 0, 0)

update_pj2(globals(), l)


opcode_arg_fmt = {
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)



# (C) Copyright 2017 by Rocky Bernstein
"""
CPython 2.0 bytecode opcodes

This is similar to the opcode portion in Python 2.0's dis.py library.
"""

import xdis.opcodes.opcode_21 as opcode_21
from xdis.opcodes.base import (
    init_opdata, finalize_opcodes, format_extended_arg, rm_op, update_pj2)

version = 2.0

l = locals()
init_opdata(l, opcode_21, version)

# 2.1 Bytecodes not in 2.0
rm_op(l, 'CONTINUE_LOOP', 119)
rm_op(l, 'MAKE_CLOSURE',  134)
rm_op(l, 'LOAD_CLOSURE',  135)
rm_op(l, 'LOAD_DEREF',    136)
rm_op(l, 'STORE_DEREF',   137)

update_pj2(globals(), l)

opcode_arg_fmt = {
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)



"""Python bytecode and instruction classes
Extracted from Python 3 dis module but generalized to
allow running on Python 2.
"""

import re, sys, types
from xdis import PYTHON3

from collections import namedtuple

from xdis.util import (get_code_object, code2num, num2code, format_code_info)

if PYTHON3:
    from io import StringIO
    from functools import reduce
else:
    from StringIO import StringIO


_have_code = (types.MethodType, types.FunctionType, types.CodeType, type)

def extended_arg_val(opc, val):
    return val << opc.EXTENDED_ARG_SHIFT

# This is modified from Python 3.6's dis
def unpack_opargs_bytecode(code, opc):
    extended_arg = 0
    try:
        n = len(code)
    except TypeError:
        code = code.co_code
        n = len(code)

    offset = 0
    while offset < n:
        prev_offset = offset
        op = code2num(code, offset)
        offset += 1
        if op_has_argument(op, opc):
            arg = code2num(code, offset) | extended_arg
            extended_arg = extended_arg_val(opc, arg) if op == opc.EXTENDED_ARG else 0
            offset += 2
        else:
            arg = None
        yield (prev_offset, op, arg)

def findlinestarts(code, dup_lines=False):
    """Find the offsets in a byte code which are start of lines in the source.

    Generate pairs (offset, lineno) as described in Python/compile.c.

    """
    if not isinstance(code.co_lnotab, str):
        byte_increments = list(code.co_lnotab[0::2])
        line_increments = list(code.co_lnotab[1::2])
    else:
        byte_increments = [ord(c) for c in code.co_lnotab[0::2]]
        line_increments = [ord(c) for c in code.co_lnotab[1::2]]

    lastlineno = None
    lineno = code.co_firstlineno
    addr = 0
    for byte_incr, line_incr in zip(byte_increments, line_increments):
        if byte_incr:
            if (lineno != lastlineno or
                (dup_lines and 0 < byte_incr < 255)):
                yield (addr, lineno)
                lastlineno = lineno
                pass
            addr += byte_incr
            pass
        lineno += line_incr
    if (lineno != lastlineno or
        (dup_lines and 0 < byte_incr < 255)):
        yield (addr, lineno)

def offset2line(offset, linestarts):
    """linestarts is expected to be a *list) of (offset, line number)
    where both offset and line number are in increasing order.
    Return the closes line number at or below the offset.
    If offset is less than the first line number given in linestarts,
    return line number 0.
    """
    if len(linestarts) == 0 or offset < linestarts[0][0]:
        return 0
    low = 0
    high = len(linestarts) - 1
    mid = (low + high + 1) // 2
    while low <= high:
        if linestarts[mid][0] > offset:
            high = mid - 1
        elif linestarts[mid][0] < offset:
            low = mid + 1
        else:
            return linestarts[mid][1]
        mid = (low + high + 1) // 2
        pass
    # Not found. Return closest position below
    if mid >= len(linestarts):
        return linestarts[len(linestarts)-1][1]
    return linestarts[high][1]

def get_jump_targets(code, opc):
    """Returns a list of instruction offsets in the supplied bytecode
    which are the targets of some sort of jump instruction.
    """
    offsets = []
    for offset, op, arg in unpack_opargs_bytecode(code, opc):
        if arg is not None:
            jump_offset = -1
            if op in opc.JREL_OPS:
                jump_offset = offset + 3 + arg
            elif op in opc.JABS_OPS:
                jump_offset = arg
            if jump_offset >= 0:
                if jump_offset not in offsets:
                    offsets.append(jump_offset)
    return offsets

def get_jump_target_maps(code, opc):
    """Returns a dictionary where the key is an offset and the values are
    a list of instruction offsets which can get run before that
    instruction. This includes jump instructions as well as non-jump
    instructions. Therefore, the keys of the dictionary are reachible
    instructions. The values of the dictionary may be useful in control-flow
    analysis.
    """
    offset2prev = {}
    prev_offset = -1
    for offset, op, arg in unpack_opargs_bytecode(code, opc):
        if prev_offset >= 0:
            prev_list = offset2prev.get(offset, [])
            prev_list.append(prev_offset)
            offset2prev[offset] = prev_list
        if op in opc.NOFOLLOW:
            prev_offset = -1
        else:
            prev_offset = offset
        if arg is not None:
            jump_offset = -1
            if op in opc.JREL_OPS:
                jump_offset = offset + 3 + arg
            elif op in opc.JABS_OPS:
                jump_offset = arg
            if jump_offset >= 0:
                prev_list = offset2prev.get(jump_offset, [])
                prev_list.append(offset)
                offset2prev[jump_offset] = prev_list
    return offset2prev


findlabels = get_jump_targets

def _get_const_info(const_index, const_list):
    """Helper to get optional details about const references

       Returns the dereferenced constant and its repr if the constant
       list is defined.
       Otherwise returns the constant index and its repr().
    """
    argval = const_index
    if const_list is not None:
        argval = const_list[const_index]

    # float values nan and inf are not directly representable in Python at least
    # before 3.5 and even there it is via a library constant.
    # So we will canonicalize their representation as float('nan') and float('inf')
    if isinstance(argval, float) and str(argval) in frozenset(['nan', '-nan', 'inf', '-inf']):
        return argval, "float('%s')" % argval
    return argval, repr(argval)

def _get_name_info(name_index, name_list):
    """Helper to get optional details about named references

       Returns the dereferenced name as both value and repr if the name
       list is defined.
       Otherwise returns the name index and its repr().
    """
    argval = name_index
    if (name_list is not None
        # PyPY seems to "optimize" out constant names,
        # so we need for that:
        and name_index < len(name_list)):
        argval = name_list[name_index]
        argrepr = argval
    else:
        argrepr = repr(argval)
    return argval, argrepr

def get_instructions_bytes(bytecode, opc, varnames=None, names=None, constants=None,
                           cells=None, linestarts=None, line_offset=0):
    """Iterate over the instructions in a bytecode string.

    Generates a sequence of Instruction namedtuples giving the details of each
    opcode.  Additional information about the code's runtime environment
    (e.g. variable names, constants) can be specified using optional
    arguments.

    """
    labels = opc.findlabels(bytecode, opc)
    extended_arg = 0

    # FIXME: We really need to distinguish 3.6.0a1 from 3.6.a3.
    # See below FIXME
    python_36 = True if opc.python_version >= 3.6 else False

    starts_line = None
    # enumerate() is not an option, since we sometimes process
    # multiple elements on a single pass through the loop
    n = len(bytecode)
    i = 0
    extended_arg = 0
    while i < n:
        op = code2num(bytecode, i)

        offset = i
        if linestarts is not None:
            starts_line = linestarts.get(i, None)
            if starts_line is not None:
                starts_line += line_offset
        is_jump_target = i in labels
        i = i+1
        arg = None
        argval = None
        argrepr = ''
        has_arg = op_has_argument(op, opc)
        optype = None
        if has_arg:
            if python_36:
                arg = code2num(bytecode, i) | extended_arg
                extended_arg = (arg << 8) if op == opc.EXTENDED_ARG else 0
                # FIXME: Python 3.6.0a1 is 2, for 3.6.a3 we have 1
                i += 1
            else:
                arg = code2num(bytecode, i) + code2num(bytecode, i+1)*256 + extended_arg
                i += 2
                extended_arg = arg*65536 if op == opc.EXTENDED_ARG else 0

            #  Set argval to the dereferenced value of the argument when
            #  availabe, and argrepr to the string representation of argval.
            #    disassemble_bytes needs the string repr of the
            #    raw name index for LOAD_GLOBAL, LOAD_CONST, etc.
            argval = arg
            if op in opc.CONST_OPS:
                argval, argrepr = _get_const_info(arg, constants)
                optype = 'const'
            elif op in opc.NAME_OPS:
                argval, argrepr = _get_name_info(arg, names)
                optype = 'name'
            elif op in opc.JREL_OPS:
                argval = i + arg
                argrepr = "to " + repr(argval)
                optype = 'jrel'
            elif op in opc.JABS_OPS:
                argval = arg
                argrepr = "to " + repr(argval)
                optype = 'jabs'
            elif op in opc.LOCAL_OPS:
                argval, argrepr = _get_name_info(arg, varnames)
                optype = 'local'
            elif op in opc.COMPARE_OPS:
                argval = opc.cmp_op[arg]
                argrepr = argval
                optype = 'compare'
            elif op in opc.FREE_OPS:
                argval, argrepr = _get_name_info(arg, cells)
                optype = 'free'
            elif op in opc.NARGS_OPS:
                optype = 'nargs'
                if not python_36:
                    argrepr = ("%d positional, %d keyword pair" %
                               (code2num(bytecode, i-2), code2num(bytecode, i-1)))
            # This has to come after hasnargs. Some are in both?
            elif op in opc.VARGS_OPS:
                optype = 'vargs'
            if hasattr(opc, 'opcode_arg_fmt') and opc.opname[op] in opc.opcode_arg_fmt:
                argrepr = opc.opcode_arg_fmt[opc.opname[op]](arg)
        elif python_36:
            i += 1

        opname = opc.opname[op]
        inst_size = op_size(op, opc)
        yield Instruction(opname, op, optype, inst_size, arg, argval, argrepr,
                          has_arg, offset, starts_line, is_jump_target)

def op_has_argument(op, opc):
    return op >= opc.HAVE_ARGUMENT

def next_offset(op, opc, offset):
    return offset + op_size(op, opc)

# FIXME: this would better be called an instr_size
# since it is about instructions, not opcodes
def instruction_size(op, opc):
    """For a given opcode, `op`, in opcode module `opc`,
    return the size, in bytes, of an `op` instruction.

    This is the size of the opcode (1 byte) and any operand it has. In
    Python before version 3.6 this will be either 1 or 3 bytes.  In
    Python 3.6 or later, it is 2 bytes or a "word"."""
    if op < opc.HAVE_ARGUMENT:
        return 2 if opc.version >= 3.6 else 1
    else:
        return 2 if opc.version >= 3.6 else 3

# Compatiblity
op_size = instruction_size


_Instruction = namedtuple("_Instruction",
     "opname opcode optype inst_size arg argval argrepr has_arg offset starts_line is_jump_target")

class Instruction(_Instruction):
    """Details for a bytecode operation

       Defined fields:
         opname - human readable name for operation
         opcode - numeric code for operation
         optype - opcode classification. One of
            compare, const, free, jabs, jrel, local, name, nargs
         inst_size - number of bytes the instruction occupies
         arg - numeric argument to operation (if any), otherwise None
         argval - resolved arg value (if known), otherwise same as arg
         argrepr - human readable description of operation argument
         has_arg - True opcode takes an argument. In that case,
                   argval and argepr will have that value. False
                   if this opcode doesn't take an argument. In that case,
                   don't look at argval or argrepr.
         offset - start index of operation within bytecode sequence
         starts_line - line started by this opcode (if any), otherwise None
         is_jump_target - True if other code jumps to here, otherwise False
    """
    # FIXME: remove has_arg from initialization but keep it as a field.

    def disassemble(self, lineno_width=3, mark_as_current=False, asm_format=False):
        """Format instruction details for inclusion in disassembly output

        *lineno_width* sets the width of the line number field (0 omits it)
        *mark_as_current* inserts a '-->' marker arrow as part of the line
        """
        fields = []
        if asm_format:
            indexed_operand = set(['name', 'local', 'compare', 'free'])
        # Column: Source code line number
        if lineno_width:
            if self.starts_line is not None:
                if asm_format:
                    lineno_fmt = "%%%dd:\n" % lineno_width
                    fields.append(lineno_fmt % self.starts_line)
                    fields.append(' ' * (lineno_width))
                    if self.is_jump_target:
                        fields.append(' ' * (lineno_width-1))
                else:
                    lineno_fmt = "%%%dd:" % lineno_width
                    fields.append(lineno_fmt % self.starts_line)
            else:
                fields.append(' ' * (lineno_width+1))
        # Column: Current instruction indicator
        if mark_as_current and not asm_format:
            fields.append('-->')
        else:
            fields.append('   ')
        # Column: Jump target marker
        if self.is_jump_target:
            if not asm_format:
                fields.append('>>')
            else:
                fields = ["L%d:\n" % self.offset] + fields
                if not self.starts_line:
                    fields.append(' ')
        else:
            fields.append('  ')
        # Column: Instruction offset from start of code sequence
        if not asm_format:
            fields.append(repr(self.offset).rjust(4))

        # Column: Opcode name
        fields.append(self.opname.ljust(20))
        # Column: Opcode argument
        if self.arg is not None:
            argrepr = self.argrepr
            if asm_format:
                if self.optype == 'jabs':
                    fields.append('L' + str(self.arg))
                elif self.optype == 'jrel':
                    argval = self.offset + self.arg + self.inst_size
                    fields.append('L' + str(argval))
                elif self.optype in indexed_operand:
                    fields.append('(%s)' % argrepr)
                    argrepr = None
                elif (self.optype == 'const'
                      and not re.search('\s', argrepr)):
                    fields.append('(%s)' % argrepr)
                    argrepr = None
                else:
                    fields.append(repr(self.arg))
            else:
                fields.append(repr(self.arg).rjust(6))
            # Column: Opcode argument details
            if argrepr:
                fields.append('(%s)' % argrepr)
                pass
            pass
        return ' '.join(fields).rstrip()

    # FIXME: figure out how to do disassemble passing in opnames

class Bytecode(object):
    """The bytecode operations of a piece of code

    Instantiate this with a function, method, string of code, or a code object
    (as returned by compile()).

    Iterating over this yields the bytecode operations as Instruction instances.
    """
    def __init__(self, x, opc, first_line=None, current_offset=None,
                 dup_lines=False):
        self.codeobj = co = get_code_object(x)
        if first_line is None:
            self.first_line = co.co_firstlineno
            self._line_offset = 0
        else:
            self.first_line = first_line
            self._line_offset = first_line - co.co_firstlineno
        self._cell_names = co.co_cellvars + co.co_freevars
        self._linestarts = dict(opc.findlinestarts(co, dup_lines=dup_lines))
        self._original_object = x
        self.opc = opc
        self.opnames = opc.opname
        self.current_offset = current_offset

    def __iter__(self):
        co = self.codeobj
        return get_instructions_bytes(co.co_code, self.opc, co.co_varnames, co.co_names,
                                      co.co_consts, self._cell_names,
                                      self._linestarts,
                                      line_offset=self._line_offset)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__,
                                 self._original_object)

    @classmethod
    def from_traceback(cls, tb):
        """ Construct a Bytecode from the given traceback """
        while tb.tb_next:
            tb = tb.tb_next
        return cls(tb.tb_frame.f_code, current_offset=tb.tb_lasti)

    def info(self):
        """Return formatted information about the code object."""
        return format_code_info(self.codeobj, self.opc.version)

    def dis(self, asm_format=False):
        """Return a formatted view of the bytecode operations."""
        co = self.codeobj
        if self.current_offset is not None:
            offset = self.current_offset
        else:
            offset = -1
        output = StringIO()
        self.disassemble_bytes(co.co_code, varnames=co.co_varnames,
                               names=co.co_names, constants=co.co_consts,
                               cells=self._cell_names,
                               linestarts=self._linestarts,
                               line_offset=self._line_offset,
                               file=output,
                               lasti=offset,
                               asm_format=asm_format)
        return output.getvalue()

    def disassemble_bytes(self, code, lasti=-1, varnames=None, names=None,
                          constants=None, cells=None, linestarts=None,
                          file=sys.stdout, line_offset=0,
                          asm_format=False):
        # Omit the line number column entirely if we have no line number info
        show_lineno = linestarts is not None
        # TODO?: Adjust width upwards if max(linestarts.values()) >= 1000?
        lineno_width = 3 if show_lineno else 0
        for instr in get_instructions_bytes(code, self.opc, varnames, names,
                                             constants, cells, linestarts,
                                             line_offset=line_offset):
            new_source_line = (show_lineno and
                               instr.starts_line is not None and
                               instr.offset > 0)
            if new_source_line:
                file.write("\n")
            is_current_instr = instr.offset == lasti
            file.write(instr.disassemble(lineno_width, is_current_instr, asm_format)
                       + "\n")
            pass
        return

    def get_instructions(self, x, first_line=None):
        """Iterator for the opcodes in methods, functions or code

        Generates a series of Instruction named tuples giving the details of
        each operations in the supplied code.

        If *first_line* is not None, it indicates the line number that should
        be reported for the first source line in the disassembled code.
        Otherwise, the source line information (if any) is taken directly from
        the disassembled code object.
        """
        co = get_code_object(x)
        cell_names = co.co_cellvars + co.co_freevars
        linestarts = dict(self.opc.findlinestarts(co))
        if first_line is not None:
            line_offset = first_line - co.co_firstlineno
        else:
            line_offset = 0
        return get_instructions_bytes(co.co_code, self.opc, co.co_varnames,
                                      co.co_names, co.co_consts, cell_names, linestarts,
                                      line_offset)

def list2bytecode(l, opc, varnames, consts):
    """Convert list/tuple of list/tuples to bytecode
    _names_ contains a list of name objects
    """
    bc = []
    for i, opcodes in enumerate(l):
        opname = opcodes[0]
        operands = opcodes[1:]
        if opname not in opc.opname:
            raise TypeError(
                "error at item %d [%s, %s], opcode not valid" %
                (i, opname, operands))
        opcode = opc.opmap[opname]
        bc.append(opcode)
        print(opname, operands)
        gen = (j for j in operands if operands)
        for j in gen:
            k = (consts if opcode in opc.CONST_OPS else varnames).index(j)
            if k == -1:
                raise TypeError(
                    "operand %s [%s, %s], not found in names" %
                    (i, opname, operands))
            else:
                bc += num2code(k)
                pass
            pass
        pass
    if opc.python_version < 3.0:
        return reduce(lambda a, b: a + chr(b), bc, '')
    else:
        if PYTHON3:
            return bytes(bc)
        else:
            return bytes(bytearray(bc))


if __name__ == '__main__':
    import xdis.opcodes.opcode_27  as opcode_27
    import xdis.opcodes.opcode_34  as opcode_34
    import xdis.opcodes.opcode_36  as opcode_36
    consts = (None, 2)
    varnames = ('a')
    instructions = [
        ('LOAD_CONST', 2),
        ('STORE_FAST', 'a'),
        ('LOAD_FAST', 'a'),
        ('RETURN_VALUE',)
    ]
    def f():
        a = 2
        return a
    if PYTHON3:
        print(f.__code__.co_code)
    else:
        print(f.func_code.co_code)

    bc = list2bytecode(instructions, opcode_27, varnames, consts)
    print(bc)
    bc = list2bytecode(instructions, opcode_34, varnames, consts)
    print(bc)
    bc = list2bytecode(instructions, opcode_36, varnames, consts)
    print(bc)



# (C) Copyright 2017 by Rocky Bernstein
"""
CPython 2.6 bytecode opcodes

This is a like Python 2.6's opcode.py with some classification
of stack usage.
"""

from xdis.opcodes.base import (
    def_op, finalize_opcodes, format_extended_arg,
    init_opdata, update_pj2)
import xdis.opcodes.opcode_25 as opcode_25

version = 2.6

l = locals()
init_opdata(l, opcode_25, version)

# Below are opcode changes since Python 2.5
def_op(l, 'STORE_MAP', 54, 3, 1)

# FIXME remove (fix uncompyle6)
update_pj2(globals(), l)

opcode_arg_fmt = {
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)



# (C) Copyright 2017 by Rocky Bernstein
"""
CPython 2.1 bytecode opcodes

This is similar to the opcode portion in Python 2.1's dis.py library.
"""

import xdis.opcodes.opcode_22 as opcode_22
from xdis.opcodes.base import (
    init_opdata, finalize_opcodes, format_extended_arg, rm_op, update_pj2)

version = 2.1

l = locals()
init_opdata(l, opcode_22, version)

# 2.1 bytecodes changes from 2.2
rm_op(l, 'BINARY_FLOOR_DIVIDE',  26)
rm_op(l, 'BINARY_TRUE_DIVIDE',   27)
rm_op(l, 'INPLACE_FLOOR_DIVIDE', 28)
rm_op(l, 'INPLACE_TRUE_DIVIDE',  29)
rm_op(l, 'GET_ITER',             68)
rm_op(l, 'YIELD_VALUE',          86)
rm_op(l, 'FOR_ITER',             93)

update_pj2(globals(), l)

opcode_arg_fmt = {
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)



#!/usr/bin/env python

import os, unittest

from xdis.load import load_module

def get_srcdir():
    filename = os.path.normcase(os.path.dirname(os.path.abspath(__file__)))
    return os.path.realpath(filename)

srcdir = get_srcdir()

class TestMarshal(unittest.TestCase):

    def test_basic(self):
        """Tests xdis.load.load_module"""
        # We deliberately pick a bytecode that we aren't likely to be running against
        mod_file = os.path.join(get_srcdir(), '..', 'test', 'bytecode_2.5',
                            '02_complex.pyc')

        (version, timestamp, magic_int, co, is_pypy,
         source_size) = load_module(mod_file)
        self.assertEqual(version, 2.5,
                         "Should have picked up Python version properly")
        assert co.co_consts == (5j, None), "Code should have a complex constant"

        mod_file = os.path.join(get_srcdir(), '..', 'test', 'bytecode_3.3',
                            '06_frozenset.pyc')
        (version, timestamp, magic_int, co, is_pypy,
         source_size) = load_module(mod_file)
        expect = (0, None, 'attlist', 'linktype', 'link', 'element', 'Yep',
                      frozenset(['linktype', 'attlist', 'element', 'link']))
        self.assertEqual(co.co_consts, expect, "Should handle frozenset")

if __name__ == '__main__':
    unittest.main()



import unittest, imp, sys
from xdis import PYTHON_VERSION, IS_PYPY
import xdis.magics as magics

class TestMagics(unittest.TestCase):

    def test_basic(self):
        """Basic test of magic numbers"""
        current = imp.get_magic()
        if hasattr(sys, 'version_info'):
            version = '.'.join([str(v) for v in sys.version_info[0:3]])
            if IS_PYPY:
                version += 'pypy'
            self.assertTrue(version in magics.magics.keys(),
                            "version %s is not in magic.magics.keys: %s" %
                            (version, magics.magics.keys()))

        self.assertEqual(current, magics.int2magic(magics.magic2int(current)))
        lookup = str(PYTHON_VERSION)
        if IS_PYPY:
            lookup += 'pypy'
        self.assertTrue(lookup in magics.magics.keys(),
                        "PYTHON VERSION %s is not in magic.magics.keys: %s" %
                        (lookup, magics.magics.keys()))

        self.assertEqual(magics.sysinfo2magic(), current,
                        "magic from imp.get_magic() should be sysinfo2magic()")


if __name__ == '__main__':
    unittest.main()



# Minimal tests for dis module

import unittest
import sys
import xdis.std as dis
import io
import re
import types
import contextlib

class _C:
    def __init__(self, x):
        self.x = x == 1

if sys.version_info[0:2] == (3, 4):
    from test.bytecode_helper import BytecodeTestCase
    from test.support import captured_stdout

    def get_tb():
        def _error():
            try:
                1 / 0
            except Exception as e:
                tb = e.__traceback__
            return tb

        tb = _error()
        while tb.tb_next:
            tb = tb.tb_next
        return tb

    TRACEBACK_CODE = get_tb().tb_frame.f_code

    dis_c_instance_method = """\
 %4d:         0 LOAD_FAST                 1 (x)
              3 LOAD_CONST                1 (1)
              6 COMPARE_OP                2 (==)
              9 LOAD_FAST                 0 (self)
             12 STORE_ATTR                0 (x)
             15 LOAD_CONST                0 (None)
             18 RETURN_VALUE

""" % (_C.__init__.__code__.co_firstlineno + 1,)

    dis_c_instance_method_bytes = """\
          0 LOAD_FAST            1 (1)
          3 LOAD_CONST           1 (1)
          6 COMPARE_OP           2 (==)
          9 LOAD_FAST            0 (0)
         12 STORE_ATTR           0 (0)
         15 LOAD_CONST           0 (0)
         18 RETURN_VALUE

"""

    def _f(a):
        print(a)
        return 1

    dis_f = """\
%3d:           0 LOAD_GLOBAL               0 (print)
                3 LOAD_FAST                 0 (a)
                6 CALL_FUNCTION             1 (1 positional, 0 keyword pair)
                9 POP_TOP

%3d:           10 LOAD_CONST                1 (1)
               13 RETURN_VALUE

""" % (_f.__code__.co_firstlineno + 1,
       _f.__code__.co_firstlineno + 2)


    dis_f_co_code = """\
 57:           0 LOAD_GLOBAL               0 (print)
                3 LOAD_FAST                 0 (a)
                6 CALL_FUNCTION             1 (1 positional, 0 keyword pair)
                9 POP_TOP

 58:          10 LOAD_CONST                1 (1)
               13 RETURN_VALUE

"""


    def bug708901():
        for res in range(1,
                         10):
            pass
        return

    dis_bug708901 = """\
%4d:         0 SETUP_LOOP               23 (to 26)
             3 LOAD_GLOBAL               0 (range)
             6 LOAD_CONST                1 (1)

%4d:         9 LOAD_CONST                2 (10)
            12 CALL_FUNCTION             2 (2 positional, 0 keyword pair)
            15 GET_ITER
       >>   16 FOR_ITER                  6 (to 25)
            19 STORE_FAST                0 (res)

%4d:        22 JUMP_ABSOLUTE            16 (to 16)
       >>   25 POP_BLOCK
       >>   26 LOAD_CONST                0 (None)
            29 RETURN_VALUE

""" % (bug708901.__code__.co_firstlineno + 1,
       bug708901.__code__.co_firstlineno + 2,
       bug708901.__code__.co_firstlineno + 3)


    def bug1333982(x=[]):
        assert 0, ([s for s in x] +
                  1)
        pass

    dis_bug1333982 = """\
%4d:         0 LOAD_CONST                1 (0)
             3 POP_JUMP_IF_TRUE         92 (to 92)
             6 LOAD_GLOBAL               0 (AssertionError)
             9 LOAD_CONST                2 (<code object <listcomp> at 0x..., file "%s", line %d>)
            12 LOAD_CONST                3 ('bug1333982.<locals>.<listcomp>')
            15 MAKE_FUNCTION             0 (0 positional, 0 name and default, 0 annotations)
            18 LOAD_FAST                 0 (x)
            21 GET_ITER
            22 CALL_FUNCTION             1 (1 positional, 0 keyword pair)

%4d:        25 LOAD_CONST                4 (1)
            28 BINARY_ADD
            29 CALL_FUNCTION             1 (1 positional, 0 keyword pair)
            32 RAISE_VARARGS             1

%4d:   >>   35 LOAD_CONST                0 (None)
            38 RETURN_VALUE

""" % (bug1333982.__code__.co_firstlineno + 1,
       __file__,
       bug1333982.__code__.co_firstlineno + 1,
       bug1333982.__code__.co_firstlineno + 2,
       bug1333982.__code__.co_firstlineno + 3)

    _BIG_LINENO_FORMAT = """\
%3d:           0 LOAD_GLOBAL               0 (spam)
               3 POP_TOP
               4 LOAD_CONST                0 (None)
               7 RETURN_VALUE

"""

    dis_module_expected_results = """\
Disassembly of f:
  4           0 LOAD_CONST                0 (None)
              3 RETURN_VALUE

Disassembly of g:
  5           0 LOAD_CONST                0 (None)
              3 RETURN_VALUE

"""

    expr_str = "x + 1"

    dis_expr_str = """\
  1:           0 LOAD_NAME                 0 (x)
               3 LOAD_CONST                0 (1)
               6 BINARY_ADD
               7 RETURN_VALUE

"""

    simple_stmt_str = "x = x + 1"

    dis_simple_stmt_str = """\
  1:           0 LOAD_NAME                 0 (x)
               3 LOAD_CONST                0 (1)
               6 BINARY_ADD
               7 STORE_NAME                0 (x)
              10 LOAD_CONST                1 (None)
              13 RETURN_VALUE

"""

    compound_stmt_str = """\
x = 0
while 1:
    x += 1"""
    # Trailing newline has been deliberately omitted

    dis_compound_stmt_str = """\
  1           0 LOAD_CONST                0 (0)
              3 STORE_NAME                0 (x)

  2           6 SETUP_LOOP               13 (to 22)

  3     >>    9 LOAD_NAME                 0 (x)
             12 LOAD_CONST                1 (1)
             15 INPLACE_ADD
             16 STORE_NAME                0 (x)
             19 JUMP_ABSOLUTE             9 (to 9)
        >>   22 LOAD_CONST                2 (None)
             25 RETURN_VALUE

"""

    dis_traceback = """\
 %-4d         0 SETUP_EXCEPT             12 (to 15)

 %-4d         3 LOAD_CONST                1 (1)
              6 LOAD_CONST                2 (0)
    -->       9 BINARY_TRUE_DIVIDE
             10 POP_TOP
             11 POP_BLOCK
             12 JUMP_FORWARD             46 (to 61)

 %-4d   >>   15 DUP_TOP
             16 LOAD_GLOBAL               0 (Exception)
             19 COMPARE_OP               10 (exception match)
             22 POP_JUMP_IF_FALSE        60
             25 POP_TOP
             26 STORE_FAST                0 (e)
             29 POP_TOP
             30 SETUP_FINALLY            14 (to 47)

 %-4d        33 LOAD_FAST                 0 (e)
             36 LOAD_ATTR                 1 (__traceback__)
             39 STORE_FAST                1 (tb)
             42 POP_BLOCK
             43 POP_EXCEPT
             44 LOAD_CONST                0 (None)
        >>   47 LOAD_CONST                0 (None)
             50 STORE_FAST                0 (e)
             53 DELETE_FAST               0 (e)
             56 END_FINALLY
             57 JUMP_FORWARD              1 (to 61)
        >>   60 END_FINALLY

 %-4d   >>   61 LOAD_FAST                1 (tb)
             64 RETURN_VALUE

""" % (TRACEBACK_CODE.co_firstlineno + 1,
       TRACEBACK_CODE.co_firstlineno + 2,
       TRACEBACK_CODE.co_firstlineno + 3,
       TRACEBACK_CODE.co_firstlineno + 4,
       TRACEBACK_CODE.co_firstlineno + 5)

    class DisTests(unittest.TestCase):


        def get_disassembly(self, func, lasti=-1, wrapper=True):
            # We want to test the default printing behaviour, not the file arg
            output = io.StringIO()
            with contextlib.redirect_stdout(output):
                if wrapper:
                    dis.dis(func)
                else:
                    dis.disassemble(func, lasti)
            return output.getvalue()

        def get_disassemble_as_string(self, func, lasti=-1):
            return self.get_disassembly(func, lasti, False)

        def strip_addresses(self, text):
            return re.sub(r'\b0x[0-9A-Fa-f]+\b', '0x...', text)

        def dis_disassembly34(self, func, expected):
            got = self.get_disassembly(func)
            if got != expected:
                got = self.strip_addresses(got)
            self.assertEqual(got, expected)

        def test_opmap(self):
            self.assertEqual(dis.opmap["NOP"], 9)
            self.assertIn(dis.opmap["LOAD_CONST"], dis.hasconst)
            self.assertIn(dis.opmap["STORE_NAME"], dis.hasname)

        def test_opname(self):
            self.assertEqual(dis.opname[dis.opmap["LOAD_FAST"]], "LOAD_FAST")

        def test_boundaries(self):
            self.assertEqual(dis.opmap["EXTENDED_ARG"], dis.EXTENDED_ARG)
            self.assertEqual(dis.opmap["STORE_NAME"], dis.HAVE_ARGUMENT)

        # def test_dis(self):
        #     self.dis_disassembly34(_f, dis_f)

        def test_bug_708901(self):
            self.skipTest('fix 708901')
            self.dis_disassembly34(bug708901, dis_bug708901)

        def test_bug_1333982(self):
            # This one is checking bytecodes generated for an `assert` statement,
            # so fails if the tests are run with -O.  Skip this test then.
            if not False:
                self.skipTest('need asserts, run without -O')

            self.dis_disassembly34(bug1333982, dis_bug1333982)

        def test_big_linenos(self):
            def func(count):
                namespace = {}
                func = "def foo():\n " + "".join(["\n "] * count + ["spam\n"])
                exec(func, namespace)
                return namespace['foo']

            # Test all small ranges
            for i in range(1, 300):
                expected = _BIG_LINENO_FORMAT % (i + 2)
                self.dis_disassembly34(func(i), expected)

            # Test some larger ranges too
            for i in range(300, 5000, 10):
                expected = _BIG_LINENO_FORMAT % (i + 2)
                self.dis_disassembly34(func(i), expected)

            self.skipTest('Add ability to disassemble module')
            # from test import dis_module
            # self.dis_disassembly34(dis_module, dis_module_expected_results)

        def test_disassemble_str(self):
            self.dis_disassembly34(expr_str, dis_expr_str)
            self.dis_disassembly34(simple_stmt_str, dis_simple_stmt_str)
            self.skipTest('Fix dis_compund_stmt_str')
            # self.dis_disassembly34(compound_stmt_str, dis_compound_stmt_str)

        def test_disassemble_bytes(self):
            self.skipTest('Fix output')
            self.dis_disassembly34(_f.__code__, dis_f_co_code)

        def test_disassemble_method(self):
            self.skipTest('Fix output')
            self.dis_disassembly34(_C(1).__init__, dis_c_instance_method)

        def test_disassemble_method_bytes(self):
            self.skipTest('Add ability to disassemble bytes')
            # method_bytecode = _C(1).__init__.__code__.co_code
            # self.dis_disassembly34(method_bytecode, dis_c_instance_method_bytes)

        def test_dis_none(self):
            try:
                del sys.last_traceback
            except AttributeError:
                pass
            self.assertRaises(TypeError, dis.dis, None)

        def test_dis_traceback(self):
            self.skipTest('Fix up ability to disassemble straceback')
            return
            try:
                del sys.last_traceback
            except AttributeError:
                pass

            try:
                1/0
            except Exception as e:
                tb = e.__traceback__
                sys.last_traceback = tb

            tb_dis = self.get_disassemble_as_string(tb.tb_frame.f_code, tb.tb_lasti)
            self.dis_disassembly34(None, tb_dis)

        def test_dis_object(self):
            self.assertRaises(TypeError, dis.dis, object())

    class DisWithFileTests(DisTests):

        # Run the tests again, using the file arg instead of print
        def get_disassembly(self, func, lasti=-1, wrapper=True):
            output = io.StringIO()
            if wrapper:
                dis.dis(func, file=output)
            else:
                dis.disassemble(func, lasti, file=output)
            return output.getvalue()



    code_info_code_info = """\
    Method Name:              code_info
    Filename:          (.*)
    Argument count:    1
    Kw-only arguments: 0
    Number of locals:  1
    Stack size:        3
    Flags:             OPTIMIZED, NEWLOCALS, NOFREE
    Constants:
       0: %r
    Names:
       0: _format_code_info
       1: _get_code_object
    Variable names:
       0: x""" % (('Formatted details of methods, functions, or code.',)
                  if sys.flags.optimize < 2 else (None,))

    # @staticmethod
    # def tricky(x, y, z=True, *args, c, d, e=[], **kwds):
    #     def f(c=c):
    #         print(x, y, z, c, d, e, f)
    #     yield x, y, z, c, d, e, f

    code_info_tricky = """\
    Method Name:              tricky
    Filename:          (.*)
    Argument count:    3
    Kw-only arguments: 3
    Number of locals:  8
    Stack size:        7
    Flags:             OPTIMIZED, NEWLOCALS, VARARGS, VARKEYWORDS, GENERATOR
    Constants:
       0: None
       1: <code object f at (.*), file "(.*)", line (.*)>
       2: 'tricky.<locals>.f'
    Variable names:
       0: x
       1: y
       2: z
       3: c
       4: d
       5: e
       6: args
       7: kwds
    Cell variables:
       0: [edfxyz]
       1: [edfxyz]
       2: [edfxyz]
       3: [edfxyz]
       4: [edfxyz]
       5: [edfxyz]"""
    # NOTE: the order of the cell variables above depends on dictionary order!

    # co_tricky_nested_f = tricky.__func__.__code__.co_consts[1]

    code_info_tricky_nested_f = """\
    Method Name:              f
    Filename:          (.*)
    Argument count:    1
    Kw-only arguments: 0
    Number of locals:  1
    Stack size:        8
    Flags:             OPTIMIZED, NEWLOCALS, NESTED
    Constants:
       0: None
    Names:
       0: print
    Variable names:
       0: c
    Free variables:
       0: [edfxyz]
       1: [edfxyz]
       2: [edfxyz]
       3: [edfxyz]
       4: [edfxyz]
       5: [edfxyz]"""

    code_info_expr_str = """\
    Name:              <module>
    Filename:          <disassembly>
    Argument count:    0
    Kw-only arguments: 0
    Number of locals:  0
    Stack size:        2
    Flags:             NOFREE
    Constants:
       0: 1
    Names:
       0: x"""

    code_info_simple_stmt_str = """\
    Name:              <module>
    Filename:          <disassembly>
    Argument count:    0
    Kw-only arguments: 0
    Number of locals:  0
    Stack size:        2
    Flags:             NOFREE
    Constants:
       0: 1
       1: None
    Names:
       0: x"""

    code_info_compound_stmt_str = """\
    Name:              <module>
    Filename:          <disassembly>
    Argument count:    0
    Kw-only arguments: 0
    Number of locals:  0
    Stack size:        2
    Flags:             NOFREE
    Constants:
       0: 0
       1: 1
       2: None
    Names:
       0: x"""

    class CodeInfoTests(unittest.TestCase):
        test_pairs = [
          # (dis.code_info, code_info_code_info),
          # (tricky, code_info_tricky),
          # (co_tricky_nested_f, code_info_tricky_nested_f),
          # (expr_str, code_info_expr_str),
          # (simple_stmt_str, code_info_simple_stmt_str),
          # (compound_stmt_str, code_info_compound_stmt_str),
        ]

        def test_code_info(self):
            self.skipTest('Reconcile differences in format')
            return
            self.maxDiff = 1000
            for x, expected in self.test_pairs:
                self.assertRegex(dis.code_info(x), expected)

        def test_show_code(self):
            self.maxDiff = 1000
            for x, expected in self.test_pairs:
                with captured_stdout() as output:
                    dis.show_code(x)
                self.assertRegex(output.getvalue(), expected+"\n")

        def test_code_info_object(self):
            self.assertRaises(TypeError, dis.code_info, object())

        def test_pretty_flags_no_flags(self):
            self.assertEqual(dis.pretty_flags(0), '0x00000000 (0x0)')


    # Fodder for instruction introspection tests
    #   Editing any of these may require recalculating the expected output
    def outer(a=1, b=2):
        def f(c=3, d=4):
            def inner(e=5, f=6):
                print(a, b, c, d, e, f)
            print(a, b, c, d)
            return inner
        print(a, b, '', 1, [], {}, "Hello world!")
        return f

    def jumpy():
        # This won't actually run (but that's OK, we only disassemble it)
        for i in range(10):
            print(i)
            if i < 4:
                continue
            if i > 6:
                break
        else:
            print("I can haz else clause?")
        while i:
            print(i)
            i -= 1
            if i > 6:
                continue
            if i < 4:
                break
        else:
            print("Who let lolcatz into this test suite?")
        try:
            1 / 0
        except ZeroDivisionError:
            print("Here we go, here we go, here we go...")
        else:
            with i as dodgy:
                print("Never reach this")
        finally:
            print("OK, now we're done")

    # End fodder for opinfo generation tests
    expected_outer_line = 1
    _line_offset = outer.__code__.co_firstlineno - 1
    code_object_f = outer.__code__.co_consts[3]
    expected_f_line = code_object_f.co_firstlineno - _line_offset
    code_object_inner = code_object_f.co_consts[3]
    expected_inner_line = code_object_inner.co_firstlineno - _line_offset
    expected_jumpy_line = 1

    # The following lines are useful to regenerate the expected results after
    # either the fodder is modified or the bytecode generation changes
    # After regeneration, update the references to code_object_f and
    # code_object_inner before rerunning the tests

    # _instructions = dis.get_instructions(outer, first_line=expected_outer_line)
    # print('expected_opinfo_outer = [\n  ',
    #       ',\n  '.join(map(str, _instructions)), ',\n]', sep='')
    # _instructions = dis.get_instructions(outer(), first_line=expected_outer_line)
    # print('expected_opinfo_f = [\n  ',
    #       ',\n  '.join(map(str, _instructions)), ',\n]', sep='')
    # _instructions = dis.get_instructions(outer()(), first_line=expected_outer_line)
    # print('expected_opinfo_inner = [\n  ',
    #       ',\n  '.join(map(str, _instructions)), ',\n]', sep='')
    # _instructions = dis.get_instructions(jumpy, first_line=expected_jumpy_line)
    # print('expected_opinfo_jumpy = [\n  ',
    #       ',\n  '.join(map(str, _instructions)), ',\n]', sep='')


    Instruction = dis.Instruction
    expected_opinfo_outer = [
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=1, argval=3, argrepr='3', has_arg=True, offset=0, starts_line=2, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=2, argval=4, argrepr='4', has_arg=True, offset=3, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CLOSURE', opcode=135, optype='free', inst_size=3, arg=0, argval='a', argrepr='a', has_arg=True, offset=6, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CLOSURE', opcode=135, optype='free', inst_size=3, arg=1, argval='b', argrepr='b', has_arg=True, offset=9, starts_line=None, is_jump_target=False),
      Instruction(opname='BUILD_TUPLE', opcode=102, optype='vargs', inst_size=3, arg=2, argval=2, argrepr='', has_arg=True, offset=12, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=3, argval=code_object_f, argrepr=repr(code_object_f), has_arg=True, offset=15, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=4, argval='outer.<locals>.f', argrepr="'outer.<locals>.f'", has_arg=True, offset=18, starts_line=None, is_jump_target=False),
      Instruction(opname='MAKE_CLOSURE', opcode=134, optype=None, inst_size=3, arg=2, argval=2, argrepr='', has_arg=True, offset=21, starts_line=None, is_jump_target=False),
      Instruction(opname='STORE_FAST', opcode=125, optype='local', inst_size=3, arg=2, argval='f', argrepr='f', has_arg=True, offset=24, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=0, argval='print', argrepr='print', has_arg=True, offset=27, starts_line=7, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=0, argval='a', argrepr='a', has_arg=True, offset=30, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=1, argval='b', argrepr='b', has_arg=True, offset=33, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=5, argval='', argrepr="''", has_arg=True, offset=36, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=6, argval=1, argrepr='1', has_arg=True, offset=39, starts_line=None, is_jump_target=False),
      Instruction(opname='BUILD_LIST', opcode=103, optype='vargs', inst_size=3, arg=0, argval=0, argrepr='', has_arg=True, offset=42, starts_line=None, is_jump_target=False),
      Instruction(opname='BUILD_MAP', opcode=105, optype='vargs', inst_size=3, arg=0, argval=0, argrepr='', has_arg=True, offset=45, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=7, argval='Hello world!', argrepr="'Hello world!'", has_arg=True, offset=48, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=7, argval=7, argrepr='7 positional, 0 keyword pair', has_arg=True, offset=51, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=54, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=2, argval='f', argrepr='f', has_arg=True, offset=55, starts_line=8, is_jump_target=False),
      Instruction(opname='RETURN_VALUE', opcode=83, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=58, starts_line=None, is_jump_target=False),
    ]

    expected_opinfo_f = [
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=1, argval=5, argrepr='5', has_arg=True, offset=0, starts_line=3, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=2, argval=6, argrepr='6', has_arg=True, offset=3, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CLOSURE', opcode=135, optype='free', inst_size=3, arg=2, argval='a', argrepr='a', has_arg=True, offset=6, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CLOSURE', opcode=135, optype='free', inst_size=3, arg=3, argval='b', argrepr='b', has_arg=True, offset=9, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CLOSURE', opcode=135, optype='free', inst_size=3, arg=0, argval='c', argrepr='c', has_arg=True, offset=12, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CLOSURE', opcode=135, optype='free', inst_size=3, arg=1, argval='d', argrepr='d', has_arg=True, offset=15, starts_line=None, is_jump_target=False),
      Instruction(opname='BUILD_TUPLE', opcode=102, optype='vargs', inst_size=3, arg=4, argval=4, argrepr='', has_arg=True, offset=18, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=3, argval=code_object_inner, argrepr=repr(code_object_inner), has_arg=True, offset=21, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=4, argval='outer.<locals>.f.<locals>.inner', argrepr="'outer.<locals>.f.<locals>.inner'", has_arg=True, offset=24, starts_line=None, is_jump_target=False),
      Instruction(opname='MAKE_CLOSURE', opcode=134, optype=None, inst_size=3, arg=2, argval=2, argrepr='', has_arg=True, offset=27, starts_line=None, is_jump_target=False),
      Instruction(opname='STORE_FAST', opcode=125, optype='local', inst_size=3, arg=2, argval='inner', argrepr='inner', has_arg=True, offset=30, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=0, argval='print', argrepr='print', has_arg=True, offset=33, starts_line=5, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=2, argval='a', argrepr='a', has_arg=True, offset=36, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=3, argval='b', argrepr='b', has_arg=True, offset=39, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=0, argval='c', argrepr='c', has_arg=True, offset=42, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=1, argval='d', argrepr='d', has_arg=True, offset=45, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=4, argval=4, argrepr='4 positional, 0 keyword pair', has_arg=True, offset=48, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=51, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=2, argval='inner', argrepr='inner', has_arg=True, offset=52, starts_line=6, is_jump_target=False),
      Instruction(opname='RETURN_VALUE', opcode=83, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=55, starts_line=None, is_jump_target=False),
    ]

    expected_opinfo_inner = [
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=0, argval='print', argrepr='print', has_arg=True, offset=0, starts_line=4, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=0, argval='a', argrepr='a', has_arg=True, offset=3, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=1, argval='b', argrepr='b', has_arg=True, offset=6, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=2, argval='c', argrepr='c', has_arg=True, offset=9, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_DEREF', opcode=136, optype='free', inst_size=3, arg=3, argval='d', argrepr='d', has_arg=True, offset=12, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='e', argrepr='e', has_arg=True, offset=15, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=1, argval='f', argrepr='f', has_arg=True, offset=18, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=6, argval=6, argrepr='6 positional, 0 keyword pair', has_arg=True, offset=21, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=24, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=0, argval=None, argrepr='None', has_arg=True, offset=25, starts_line=None, is_jump_target=False),
      Instruction(opname='RETURN_VALUE', opcode=83, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=28, starts_line=None, is_jump_target=False),
    ]

    expected_opinfo_jumpy = [
      Instruction(opname='SETUP_LOOP', opcode=120, optype='jrel', inst_size=3, arg=74, argval=77, argrepr='to 77', has_arg=True, offset=0, starts_line=3, is_jump_target=False),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=0, argval='range', argrepr='range', has_arg=True, offset=3, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=1, argval=10, argrepr='10', has_arg=True, offset=6, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=1, argval=1, argrepr='1 positional, 0 keyword pair', has_arg=True, offset=9, starts_line=None, is_jump_target=False),
      Instruction(opname='GET_ITER', opcode=68, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=12, starts_line=None, is_jump_target=False),
      Instruction(opname='FOR_ITER', opcode=93, optype='jrel', inst_size=3, arg=50, argval=66, argrepr='to 66', has_arg=True, offset=13, starts_line=None, is_jump_target=True),
      Instruction(opname='STORE_FAST', opcode=125, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=16, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=1, argval='print', argrepr='print', has_arg=True, offset=19, starts_line=4, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=22, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=1, argval=1, argrepr='1 positional, 0 keyword pair', has_arg=True, offset=25, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=28, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=29, starts_line=5, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=2, argval=4, argrepr='4', has_arg=True, offset=32, starts_line=None, is_jump_target=False),
      Instruction(opname='COMPARE_OP', opcode=107, optype='compare', inst_size=3, arg=0, argval='<', argrepr='<', has_arg=True, offset=35, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_JUMP_IF_FALSE', opcode=114, optype='jabs', inst_size=3, arg=47, argval=47, argrepr='to 47', has_arg=True, offset=38, starts_line=None, is_jump_target=False),
      Instruction(opname='JUMP_ABSOLUTE', opcode=113, optype='jabs', inst_size=3, arg=13, argval=13, argrepr='to 13', has_arg=True, offset=41, starts_line=6, is_jump_target=False),
      Instruction(opname='JUMP_FORWARD', opcode=110, optype='jrel', inst_size=3, arg=0, argval=47, argrepr='to 47', has_arg=True, offset=44, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=47, starts_line=7, is_jump_target=True),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=3, argval=6, argrepr='6', has_arg=True, offset=50, starts_line=None, is_jump_target=False),
      Instruction(opname='COMPARE_OP', opcode=107, optype='compare', inst_size=3, arg=4, argval='>', argrepr='>', has_arg=True, offset=53, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_JUMP_IF_FALSE', opcode=114, optype='jabs', inst_size=3, arg=13, argval=13, argrepr='to 13', has_arg=True, offset=56, starts_line=None, is_jump_target=False),
      Instruction(opname='BREAK_LOOP', opcode=80, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=59, starts_line=8, is_jump_target=False),
      Instruction(opname='JUMP_ABSOLUTE', opcode=113, optype='jabs', inst_size=3, arg=13, argval=13, argrepr='to 13', has_arg=True, offset=60, starts_line=None, is_jump_target=False),
      Instruction(opname='JUMP_ABSOLUTE', opcode=113, optype='jabs', inst_size=3, arg=13, argval=13, argrepr='to 13', has_arg=True, offset=63, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_BLOCK', opcode=87, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=66, starts_line=None, is_jump_target=True),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=1, argval='print', argrepr='print', has_arg=True, offset=67, starts_line=10, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=4, argval='I can haz else clause?', argrepr="'I can haz else clause?'", has_arg=True, offset=70, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=1, argval=1, argrepr='1 positional, 0 keyword pair', has_arg=True, offset=73, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=76, starts_line=None, is_jump_target=False),
      Instruction(opname='SETUP_LOOP', opcode=120, optype='jrel', inst_size=3, arg=74, argval=154, argrepr='to 154', has_arg=True, offset=77, starts_line=11, is_jump_target=True),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=80, starts_line=None, is_jump_target=True),
      Instruction(opname='POP_JUMP_IF_FALSE', opcode=114, optype='jabs', inst_size=3, arg=143, argval=143, argrepr='to 143', has_arg=True, offset=83, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=1, argval='print', argrepr='print', has_arg=True, offset=86, starts_line=12, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=89, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=1, argval=1, argrepr='1 positional, 0 keyword pair', has_arg=True, offset=92, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=95, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=96, starts_line=13, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=5, argval=1, argrepr='1', has_arg=True, offset=99, starts_line=None, is_jump_target=False),
      Instruction(opname='INPLACE_SUBTRACT', opcode=56, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=102, starts_line=None, is_jump_target=False),
      Instruction(opname='STORE_FAST', opcode=125, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=103, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=106, starts_line=14, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=3, argval=6, argrepr='6', has_arg=True, offset=109, starts_line=None, is_jump_target=False),
      Instruction(opname='COMPARE_OP', opcode=107, optype='compare', inst_size=3, arg=4, argval='>', argrepr='>', has_arg=True, offset=112, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_JUMP_IF_FALSE', opcode=114, optype='jabs', inst_size=3, arg=124, argval=124, argrepr='to 124', has_arg=True, offset=115, starts_line=None, is_jump_target=False),
      Instruction(opname='JUMP_ABSOLUTE', opcode=113, optype='jabs', inst_size=3, arg=80, argval=80, argrepr='to 80', has_arg=True, offset=118, starts_line=15, is_jump_target=False),
      Instruction(opname='JUMP_FORWARD', opcode=110, optype='jrel', inst_size=3, arg=0, argval=124, argrepr='to 124', has_arg=True, offset=121, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=124, starts_line=16, is_jump_target=True),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=2, argval=4, argrepr='4', has_arg=True, offset=127, starts_line=None, is_jump_target=False),
      Instruction(opname='COMPARE_OP', opcode=107, optype='compare', inst_size=3, arg=0, argval='<', argrepr='<', has_arg=True, offset=130, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_JUMP_IF_FALSE', opcode=114, optype='jabs', inst_size=3, arg=80, argval=80, argrepr='to 80', has_arg=True, offset=133, starts_line=None, is_jump_target=False),
      Instruction(opname='BREAK_LOOP', opcode=80, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=136, starts_line=17, is_jump_target=False),
      Instruction(opname='JUMP_ABSOLUTE', opcode=113, optype='jabs', inst_size=3, arg=80, argval=80, argrepr='to 80', has_arg=True, offset=137, starts_line=None, is_jump_target=False),
      Instruction(opname='JUMP_ABSOLUTE', opcode=113, optype='jabs', inst_size=3, arg=80, argval=80, argrepr='to 80', has_arg=True, offset=140, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_BLOCK', opcode=87, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=143, starts_line=None, is_jump_target=True),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=1, argval='print', argrepr='print', has_arg=True, offset=144, starts_line=19, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=6, argval='Who let lolcatz into this test suite?', argrepr="'Who let lolcatz into this test suite?'", has_arg=True, offset=147, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=1, argval=1, argrepr='1 positional, 0 keyword pair', has_arg=True, offset=150, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=153, starts_line=None, is_jump_target=False),
      Instruction(opname='SETUP_FINALLY', opcode=122, optype='jrel', inst_size=3, arg=72, argval=229, argrepr='to 229', has_arg=True, offset=154, starts_line=20, is_jump_target=True),
      Instruction(opname='SETUP_EXCEPT', opcode=121, optype='jrel', inst_size=3, arg=12, argval=172, argrepr='to 172', has_arg=True, offset=157, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=5, argval=1, argrepr='1', has_arg=True, offset=160, starts_line=21, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=7, argval=0, argrepr='0', has_arg=True, offset=163, starts_line=None, is_jump_target=False),
      Instruction(opname='BINARY_TRUE_DIVIDE', opcode=27, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=166, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=167, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_BLOCK', opcode=87, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=168, starts_line=None, is_jump_target=False),
      Instruction(opname='JUMP_FORWARD', opcode=110, optype='jrel', inst_size=3, arg=28, argval=200, argrepr='to 200', has_arg=True, offset=169, starts_line=None, is_jump_target=False),
      Instruction(opname='DUP_TOP', opcode=4, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=172, starts_line=22, is_jump_target=True),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=2, argval='ZeroDivisionError', argrepr='ZeroDivisionError', has_arg=True, offset=173, starts_line=None, is_jump_target=False),
      Instruction(opname='COMPARE_OP', opcode=107, optype='compare', inst_size=3, arg=10, argval='exception match', argrepr='exception match', has_arg=True, offset=176, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_JUMP_IF_FALSE', opcode=114, optype='jabs', inst_size=3, arg=199, argval=199, argrepr='to 199', has_arg=True, offset=179, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=182, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=183, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=184, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=1, argval='print', argrepr='print', has_arg=True, offset=185, starts_line=23, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=8, argval='Here we go, here we go, here we go...', argrepr="'Here we go, here we go, here we go...'", has_arg=True, offset=188, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=1, argval=1, argrepr='1 positional, 0 keyword pair', has_arg=True, offset=191, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=194, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_EXCEPT', opcode=89, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=195, starts_line=None, is_jump_target=False),
      Instruction(opname='JUMP_FORWARD', opcode=110, optype='jrel', inst_size=3, arg=26, argval=225, argrepr='to 225', has_arg=True, offset=196, starts_line=None, is_jump_target=False),
      Instruction(opname='END_FINALLY', opcode=88, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=199, starts_line=None, is_jump_target=True),
      Instruction(opname='LOAD_FAST', opcode=124, optype='local', inst_size=3, arg=0, argval='i', argrepr='i', has_arg=True, offset=200, starts_line=25, is_jump_target=True),
      Instruction(opname='SETUP_WITH', opcode=143, optype='jrel', inst_size=3, arg=17, argval=223, argrepr='to 223', has_arg=True, offset=203, starts_line=None, is_jump_target=False),
      Instruction(opname='STORE_FAST', opcode=125, optype='local', inst_size=3, arg=1, argval='dodgy', argrepr='dodgy', has_arg=True, offset=206, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=1, argval='print', argrepr='print', has_arg=True, offset=209, starts_line=26, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=9, argval='Never reach this', argrepr="'Never reach this'", has_arg=True, offset=212, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=1, argval=1, argrepr='1 positional, 0 keyword pair', has_arg=True, offset=215, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=218, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_BLOCK', opcode=87, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=219, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=0, argval=None, argrepr='None', has_arg=True, offset=220, starts_line=None, is_jump_target=False),
      Instruction(opname='WITH_CLEANUP', opcode=81, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=223, starts_line=None, is_jump_target=True),
      Instruction(opname='END_FINALLY', opcode=88, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=224, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_BLOCK', opcode=87, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=225, starts_line=None, is_jump_target=True),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=0, argval=None, argrepr='None', has_arg=True, offset=226, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_GLOBAL', opcode=116, optype='name', inst_size=3, arg=1, argval='print', argrepr='print', has_arg=True, offset=229, starts_line=28, is_jump_target=True),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=10, argval="OK, now we're done", argrepr='"OK, now we\'re done"', has_arg=True, offset=232, starts_line=None, is_jump_target=False),
      Instruction(opname='CALL_FUNCTION', opcode=131, optype='nargs', inst_size=3, arg=1, argval=1, argrepr='1 positional, 0 keyword pair', has_arg=True, offset=235, starts_line=None, is_jump_target=False),
      Instruction(opname='POP_TOP', opcode=1, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=238, starts_line=None, is_jump_target=False),
      Instruction(opname='END_FINALLY', opcode=88, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=239, starts_line=None, is_jump_target=False),
      Instruction(opname='LOAD_CONST', opcode=100, optype='const', inst_size=3, arg=0, argval=None, argrepr='None', has_arg=True, offset=240, starts_line=None, is_jump_target=False),
      Instruction(opname='RETURN_VALUE', opcode=83, optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=243, starts_line=None, is_jump_target=False),
    ]


    # One last piece of inspect fodder to check the default line number handling
    def simple(): pass
    expected_opinfo_simple = [
      Instruction(opname='LOAD_CONST',   opcode=100, optype='const', inst_size= 3, arg=0,    argval=None, argrepr='None', has_arg=True, offset=0, starts_line=simple.__code__.co_firstlineno, is_jump_target=False),
      Instruction(opname='RETURN_VALUE', opcode=83,  optype=None, inst_size=1, arg=None, argval=None, argrepr='', has_arg=False, offset=3, starts_line=None, is_jump_target=False)
    ]


    class InstructionTests(BytecodeTestCase):

        def test_default_first_line(self):
            actual = dis.get_instructions(simple)
            self.assertEqual(list(actual), expected_opinfo_simple)

        def test_first_line_set_to_None(self):
            actual = dis.get_instructions(simple, first_line=None)
            self.assertEqual(list(actual), expected_opinfo_simple)

        def test_outer(self):
            actual = dis.get_instructions(outer, first_line=expected_outer_line)
            self.assertEqual(list(actual), expected_opinfo_outer)

        def test_nested(self):
            with captured_stdout():
                f = outer()
            actual = dis.get_instructions(f, first_line=expected_f_line)
            self.assertEqual(list(actual), expected_opinfo_f)

        def test_doubly_nested(self):
            with captured_stdout():
                inner = outer()()
            actual = dis.get_instructions(inner, first_line=expected_inner_line)
            self.assertEqual(list(actual), expected_opinfo_inner)

        # def test_jumpy(self):
        #     actual = dis.get_instructions(jumpy, first_line=expected_jumpy_line)
        #     self.assertEqual(list(actual), expected_opinfo_jumpy)

    # get_instructions has its own tests above, so can rely on it to validate
    # the object oriented API
    class BytecodeTests(unittest.TestCase):
        def test_instantiation(self):
            # Test with function, method, code string and code object
            for obj in [_f, _C(1).__init__, "a=1", _f.__code__]:
                with self.subTest(obj=obj):
                    b = dis.Bytecode(obj)
                    self.assertIsInstance(b.codeobj, types.CodeType)

            self.assertRaises(TypeError, dis.Bytecode, object())

        def test_iteration(self):
            for obj in [_f, _C(1).__init__, "a=1", _f.__code__]:
                with self.subTest(obj=obj):
                    via_object = list(dis.Bytecode(obj))
                    via_generator = list(dis.get_instructions(obj))
                    self.assertEqual(via_object, via_generator)

        def test_explicit_first_line(self):
            actual = dis.Bytecode(outer, first_line=expected_outer_line)
            self.assertEqual(list(actual), expected_opinfo_outer)

        def test_source_line_in_disassembly(self):
            # Use the line in the source code
            actual = dis.Bytecode(simple).dis()[:3]
            expected = "{:>3}".format(simple.__code__.co_firstlineno)
            self.assertEqual(actual, expected)
            # Use an explicit first line number
            actual = dis.Bytecode(simple, first_line=350).dis()[:3]
            self.assertEqual(actual, "350")

        def test_info(self):
            self.maxDiff = 1000
            for x, expected in CodeInfoTests.test_pairs:
                b = dis.Bytecode(x)
                self.assertRegex(b.info(), expected)

        def test_disassembled(self):
            actual = dis.Bytecode(_f).dis()
            self.skipTest('fix test_disassembled')
            self.assertEqual(actual, dis_f)

        def test_from_traceback(self):
            tb = get_tb()
            b = dis.Bytecode.from_traceback(tb)
            while tb.tb_next: tb = tb.tb_next

            self.assertEqual(b.current_offset, tb.tb_lasti)

        def test_from_traceback_dis(self):
            tb = get_tb()
            b = dis.Bytecode.from_traceback(tb)
            self.skipTest('fix test_from_traceback_dis')
            self.assertEqual(b.dis(), dis_traceback)

if __name__ == "__main__":
    unittest.main()
